import urllib.parse
import warnings
from abc import ABC
from collections.abc import Callable, Iterable, Iterator
from functools import lru_cache
from typing import cast

from rdflib import RDF, Graph, Literal, Namespace, URIRef
from rdflib.query import ResultRow

from cognite.neat._constants import CLASSIC_CDF_NAMESPACE, DEFAULT_NAMESPACE
from cognite.neat._graph import extractors
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings import ResourceNotFoundWarning
from cognite.neat._utils.collection_ import iterate_progress_bar
from cognite.neat._utils.rdf_ import (
    Triple,
    add_triples_in_batch,
    get_namespace,
    remove_instance_ids_in_batch,
    remove_namespace_from_uri,
)

from ._base import BaseTransformer, BaseTransformerStandardised, RowTransformationOutput


class AddAssetDepth(BaseTransformerStandardised):
    description: str = "Adds depth of asset in the asset hierarchy and optionally types asset based on depth"
    _use_only_once: bool = True
    _need_changes = frozenset({str(extractors.AssetsExtractor.__name__)})

    def __init__(
        self,
        asset_type: URIRef | None = None,
        parent_prop: URIRef | None = None,
        depth_typing: dict[int, str] | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.parent_prop = parent_prop or DEFAULT_NAMESPACE.parentId
        self.depth_typing = depth_typing

    def _iterate_query(self) -> str:
        query = """SELECT ?asset (IF(?isRoot, 0, COUNT(?parent)) AS ?parentCount)
                   WHERE {{
                        ?asset a <{asset_type}> .
                        OPTIONAL {{ ?asset <{parent_prop}>+ ?parent . }}
                        BIND(IF(BOUND(?parent), false, true) AS ?isRoot)}}
                   GROUP BY ?asset ?isRoot
                   ORDER BY DESC(?parentCount)"""

        return query.format(
            asset_type=self.asset_type,
            parent_prop=self.parent_prop,
        )

    def _count_query(self) -> str:
        query = """SELECT (COUNT(?asset) as ?count)
                   WHERE {{ ?asset a <{asset_type}> . }}"""

        return query.format(asset_type=self.asset_type)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()
        subject, object = query_result_row

        row_output.add_triples.add(cast(Triple, (subject, DEFAULT_NAMESPACE.depth, object)))

        if self.depth_typing and (type_ := self.depth_typing.get(int(object), None)):
            row_output.remove_triples.add(cast(Triple, (subject, RDF.type, self.asset_type)))
            row_output.add_triples.add(cast(Triple, (subject, RDF.type, DEFAULT_NAMESPACE[type_])))

        row_output.instances_modified_count += 1

        return row_output


class BaseAssetConnector(BaseTransformerStandardised, ABC):
    description: str = "Connects assets to other cognite resources, thus forming bi-directional connection"
    _use_only_once: bool = True

    def _count_query(self) -> str:
        query = """SELECT (COUNT(?asset) as ?count)
                   WHERE {{
                        ?resource a <{resource_type}> .
                        ?resource <{connection}> ?asset .
                        ?asset a <{asset_type}> .
                        }}"""

        return query.format(
            asset_type=self.asset_type,
            resource_type=self.resource_type,
            connection=self.resource_to_asset_connection,
        )

    def _iterate_query(self) -> str:
        query = """SELECT ?asset ?resource
                   WHERE {{
                        ?resource a <{resource_type}> .
                        ?resource <{connection}> ?asset .
                        ?asset a <{asset_type}> .
                        }}"""

        return query.format(
            asset_type=self.asset_type,
            resource_type=self.resource_type,
            connection=self.resource_to_asset_connection,
        )

    def __init__(
        self,
        resource_to_asset_connection: URIRef,
        resource_type: URIRef,
        asset_to_resource_connection: URIRef | None = None,
        asset_type: URIRef | None = None,
    ) -> None:
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.resource_to_asset_connection = resource_to_asset_connection
        self.resource_type = resource_type

        if asset_to_resource_connection:
            self.asset_to_resource_connection = asset_to_resource_connection
        else:
            namespace = Namespace(get_namespace(resource_type))
            type_ = remove_namespace_from_uri(resource_type)
            self.asset_to_resource_connection = namespace[type_[0].lower() + type_[1:]]

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()
        subject, object = query_result_row

        row_output.add_triples.add(cast(Triple, (subject, self.asset_to_resource_connection, object)))

        row_output.instances_modified_count += 1

        return row_output


class AssetTimeSeriesConnector(BaseAssetConnector):
    description: str = "Connects assets to timeseries, thus forming bi-directional connection"
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.TimeSeriesExtractor.__name__),
        }
    )

    def __init__(
        self,
        resource_to_asset_connection: URIRef | None = None,
        resource_type: URIRef | None = None,
        asset_to_resource_connection: URIRef | None = None,
        asset_type: URIRef | None = None,
    ):
        super().__init__(
            resource_to_asset_connection=resource_to_asset_connection or DEFAULT_NAMESPACE.assetId,
            resource_type=resource_type or DEFAULT_NAMESPACE.TimeSeries,
            asset_to_resource_connection=asset_to_resource_connection or DEFAULT_NAMESPACE.timeSeries,
            asset_type=asset_type or DEFAULT_NAMESPACE.Asset,
        )


class AssetSequenceConnector(BaseAssetConnector):
    description: str = "Connects assets to sequences, thus forming bi-directional connection"
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.SequencesExtractor.__name__),
        }
    )

    def __init__(
        self,
        resource_to_asset_connection: URIRef | None = None,
        resource_type: URIRef | None = None,
        asset_to_resource_connection: URIRef | None = None,
        asset_type: URIRef | None = None,
    ):
        super().__init__(
            resource_to_asset_connection=resource_to_asset_connection or DEFAULT_NAMESPACE.assetId,
            resource_type=resource_type or DEFAULT_NAMESPACE.Sequence,
            asset_to_resource_connection=asset_to_resource_connection or DEFAULT_NAMESPACE.sequence,
            asset_type=asset_type or DEFAULT_NAMESPACE.Asset,
        )


class AssetFileConnector(BaseAssetConnector):
    description: str = "Connects assets to files, thus forming bi-directional connection"
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.FilesExtractor.__name__),
        }
    )

    def __init__(
        self,
        resource_to_asset_connection: URIRef | None = None,
        resource_type: URIRef | None = None,
        asset_to_resource_connection: URIRef | None = None,
        asset_type: URIRef | None = None,
    ):
        super().__init__(
            resource_to_asset_connection=resource_to_asset_connection or DEFAULT_NAMESPACE.assetIds,
            resource_type=resource_type or DEFAULT_NAMESPACE.File,
            asset_to_resource_connection=asset_to_resource_connection or DEFAULT_NAMESPACE.file,
            asset_type=asset_type or DEFAULT_NAMESPACE.Asset,
        )


class AssetEventConnector(BaseAssetConnector):
    description: str = "Connects assets to events, thus forming bi-directional connection"
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.EventsExtractor.__name__),
        }
    )

    def __init__(
        self,
        resource_to_asset_connection: URIRef | None = None,
        resource_type: URIRef | None = None,
        asset_to_resource_connection: URIRef | None = None,
        asset_type: URIRef | None = None,
    ):
        super().__init__(
            resource_to_asset_connection=resource_to_asset_connection or DEFAULT_NAMESPACE.assetIds,
            resource_type=resource_type or DEFAULT_NAMESPACE.Event,
            asset_to_resource_connection=asset_to_resource_connection or DEFAULT_NAMESPACE.event,
            asset_type=asset_type or DEFAULT_NAMESPACE.Asset,
        )


class AssetRelationshipConnector(BaseTransformerStandardised):
    description: str = "Connects assets via relationships"
    _use_only_once: bool = True
    _need_changes = frozenset(
        {
            str(extractors.AssetsExtractor.__name__),
            str(extractors.RelationshipsExtractor.__name__),
        }
    )

    def _count_query(self) -> str:
        query = """SELECT (COUNT(?target_xid) as ?count) WHERE {{
                   ?relationship a <{relationship_type}> .
                   ?relationship <{relationship_source_xid_prop}> ?source_xid .
                   ?source_xid a <{asset_type}> .

                   ?relationship <{relationship_target_xid_prop}> ?target_xid .
                   ?target_xid a <{asset_type}> .}}"""

        return query.format(
            relationship_type=self.relationship_type,
            relationship_source_xid_prop=self.relationship_source_xid_prop,
            relationship_target_xid_prop=self.relationship_target_xid_prop,
            asset_type=self.asset_type,
        )

    def _iterate_query(self) -> str:
        query = """SELECT ?source_xid ?relationship ?target_xid WHERE {{
                   ?relationship a <{relationship_type}> .
                   ?relationship <{relationship_source_xid_prop}> ?source_xid .
                   ?source_xid a <{asset_type}> .

                   ?relationship <{relationship_target_xid_prop}> ?target_xid .
                   ?target_xid a <{asset_type}> .}}"""

        return query.format(
            relationship_type=self.relationship_type,
            relationship_source_xid_prop=self.relationship_source_xid_prop,
            relationship_target_xid_prop=self.relationship_target_xid_prop,
            asset_type=self.asset_type,
        )

    def __init__(
        self,
        asset_type: URIRef | None = None,
        relationship_type: URIRef | None = None,
        relationship_source_xid_prop: URIRef | None = None,
        relationship_target_xid_prop: URIRef | None = None,
    ):
        self.asset_type = asset_type or DEFAULT_NAMESPACE.Asset
        self.relationship_type = relationship_type or DEFAULT_NAMESPACE.Relationship
        self.relationship_source_xid_prop = relationship_source_xid_prop or DEFAULT_NAMESPACE.sourceExternalId
        self.relationship_target_xid_prop = relationship_target_xid_prop or DEFAULT_NAMESPACE.targetExternalId

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()
        source, relationship, target = query_result_row

        row_output.add_triples.add(cast(Triple, (source, DEFAULT_NAMESPACE.relationship, target)))
        row_output.add_triples.add(cast(Triple, (relationship, DEFAULT_NAMESPACE.source, source)))
        row_output.add_triples.add(cast(Triple, (relationship, DEFAULT_NAMESPACE.target, target)))

        row_output.remove_triples.add(cast(Triple, (relationship, self.relationship_source_xid_prop, None)))
        row_output.remove_triples.add(cast(Triple, (relationship, self.relationship_target_xid_prop, None)))

        row_output.instances_modified_count += 2

        return row_output


# TODO: standardise
class RelationshipAsEdgeTransformer(BaseTransformer):
    """Converts relationships into edges in the graph.

    This transformer converts relationships into edges in the graph. This is useful as the
    edges will be picked up as part of the schema connected to Assets, Events, Files, Sequenses,
    and TimeSeries in the InferenceImporter.

    Args:
        min_relationship_types: The minimum number of relationship types that must exists to convert those
            relationships to edges. For example, if there is only 5 relationships between Assets and TimeSeries,
            and limit is 10, those relationships will not be converted to edges.
        limit_per_type: The number of conversions to perform per relationship type. For example, if there are 10
            relationships between Assets and TimeSeries, and limit_per_type is 1, only 1 of those relationships
            will be converted to an edge. If None, all relationships will be converted.

    """

    def __init__(
        self,
        min_relationship_types: int = 1,
        limit_per_type: int | None = None,
        namespace: Namespace = CLASSIC_CDF_NAMESPACE,
    ) -> None:
        self._min_relationship_types = min_relationship_types
        self._limit_per_type = limit_per_type
        self._namespace = namespace

    _NOT_PROPERTIES: frozenset[str] = frozenset(
        {"sourceExternalId", "targetExternalId", "externalId", "sourceType", "targetType"}
    )
    _RELATIONSHIP_NODE_TYPES: tuple[str, ...] = tuple(["Asset", "Event", "File", "Sequence", "TimeSeries"])
    description = "Converts relationships to edge"
    _use_only_once: bool = True
    _need_changes = frozenset({extractors.RelationshipsExtractor.__name__})

    _count_by_source_target = """PREFIX classic: <{namespace}>

SELECT (COUNT(?instance) AS ?instanceCount)
WHERE {{
  ?instance a classic:Relationship .
  ?instance classic:sourceType classic:{source_type} .
  ?instance classic:targetType classic:{target_type} .
}}"""

    _instances = """PREFIX classic: <{namespace}>

SELECT ?instance
WHERE {{
    ?instance a classic:Relationship .
    ?instance classic:sourceType classic:{source_type} .
    ?instance classic:targetType classic:{target_type} .
}}"""
    _lookup_entity_query = """PREFIX classic: <{namespace}>

SELECT ?entity
WHERE {{
    ?entity a classic:{entity_type} .
    ?entity classic:externalId "{external_id}" .
}}"""

    @staticmethod
    def create_lookup_entity_with_external_id(graph: Graph, namespace: Namespace) -> Callable[[str, str], URIRef]:
        @lru_cache(maxsize=10_000)
        def lookup_entity_with_external_id(entity_type: str, external_id: str) -> URIRef:
            query = RelationshipAsEdgeTransformer._lookup_entity_query.format(
                namespace=namespace, entity_type=entity_type, external_id=external_id
            )
            result = list(graph.query(query))
            if len(result) == 1:
                return cast(URIRef, result[0][0])  # type: ignore[index]
            raise ValueError(f"Could not find entity with external_id {external_id} and type {entity_type}")

        return lookup_entity_with_external_id

    def transform(self, graph: Graph) -> None:
        lookup_entity_with_external_id = self.create_lookup_entity_with_external_id(graph, self._namespace)
        for source_type in self._RELATIONSHIP_NODE_TYPES:
            for target_type in self._RELATIONSHIP_NODE_TYPES:
                query = self._count_by_source_target.format(
                    namespace=self._namespace, source_type=source_type, target_type=target_type
                )
                for instance_count_res in graph.query(query):
                    instance_count = int(instance_count_res[0])  # type: ignore[index, arg-type]
                    if instance_count < self._min_relationship_types:
                        continue
                    edge_triples = self._edge_triples(
                        graph, source_type, target_type, instance_count, lookup_entity_with_external_id
                    )
                    add_triples_in_batch(graph, edge_triples)

    def _edge_triples(
        self,
        graph: Graph,
        source_type: str,
        target_type: str,
        instance_count: int,
        lookup_entity_with_external_id: Callable[[str, str], URIRef],
    ) -> Iterable[Triple]:
        query = self._instances.format(namespace=self._namespace, source_type=source_type, target_type=target_type)
        total_instance_count = instance_count if self._limit_per_type is None else self._limit_per_type

        converted_relationships: list[URIRef] = []
        for no, result in enumerate(
            iterate_progress_bar(graph.query(query), total=total_instance_count, description="Relationships to edges")
        ):
            if self._limit_per_type is not None and no >= self._limit_per_type:
                break
            relationship_id = cast(URIRef, result[0])  # type: ignore[index, misc]
            yield from self._relationship_as_edge(
                graph, relationship_id, source_type, target_type, lookup_entity_with_external_id
            )
            converted_relationships.append(relationship_id)

            if len(converted_relationships) >= 1_000:
                remove_instance_ids_in_batch(graph, converted_relationships)
                converted_relationships = []

        remove_instance_ids_in_batch(graph, converted_relationships)

    def _relationship_as_edge(
        self,
        graph: Graph,
        relationship_id: URIRef,
        source_type: str,
        target_type: str,
        lookup_entity_with_external_id: Callable[[str, str], URIRef],
    ) -> list[Triple]:
        relationship_triples = cast(list[Triple], list(graph.query(f"DESCRIBE <{relationship_id}>")))
        object_by_predicates = cast(
            dict[str, URIRef | Literal],
            {remove_namespace_from_uri(row[1]): row[2] for row in relationship_triples if row[1] != RDF.type},
        )
        source_external_id = cast(URIRef, object_by_predicates["sourceExternalId"])
        target_source_id = cast(URIRef, object_by_predicates["targetExternalId"])
        try:
            source_id = lookup_entity_with_external_id(source_type, source_external_id)
        except ValueError:
            warnings.warn(
                ResourceNotFoundWarning(source_external_id, "class", str(relationship_id), "class"), stacklevel=2
            )
            return []
        try:
            target_id = lookup_entity_with_external_id(target_type, target_source_id)
        except ValueError:
            warnings.warn(
                ResourceNotFoundWarning(target_source_id, "class", str(relationship_id), "class"), stacklevel=2
            )
            return []
        edge_id = urllib.parse.quote(str(object_by_predicates["externalId"]))
        # If there is properties on the relationship, we create a new intermediate node
        edge_type = self._namespace[f"{source_type}To{target_type}Edge"]
        return self._create_edge(
            object_by_predicates, edge_id, source_id, target_id, self._predicate(target_type), edge_type
        )

    def _lookup_entity(self, graph: Graph, entity_type: str, external_id: str) -> URIRef:
        query = self._lookup_entity_query.format(
            namespace=self._namespace, entity_type=entity_type, external_id=external_id
        )
        result = list(graph.query(query))
        if len(result) == 1:
            return cast(URIRef, result[0][0])  # type: ignore[index]
        raise ValueError(f"Could not find entity with external_id {external_id} and type {entity_type}")

    def _create_edge(
        self,
        objects_by_predicates: dict[str, URIRef | Literal],
        external_id: str,
        source_id: URIRef,
        target_id: URIRef,
        predicate: URIRef,
        edge_type: URIRef,
    ) -> list[Triple]:
        """Creates a new intermediate node for the relationship with properties."""
        # Create the entity with the properties
        edge_triples: list[Triple] = []
        edge_id = self._namespace[external_id]

        edge_triples.append((edge_id, RDF.type, edge_type))
        for prop_name, object_ in objects_by_predicates.items():
            if prop_name in self._NOT_PROPERTIES:
                continue
            edge_triples.append((edge_id, self._namespace[prop_name], object_))

        # Target and Source IDs will always be a combination of Asset, Sequence, Event, TimeSeries, and File.
        # If we assume source ID is an asset and target ID is a time series, then
        # before we had relationship pointing to both: timeseries <- relationship -> asset
        # After, we want asset <-> Edge -> TimeSeries
        # and the new edge will point to the asset and the timeseries through startNode and endNode

        # Link the source to the new edge
        edge_triples.append((source_id, predicate, edge_id))
        # Link the edge to the source and target
        edge_triples.append((edge_id, self._namespace["startNode"], source_id))
        edge_triples.append((edge_id, self._namespace["endNode"], target_id))
        return edge_triples

    def _predicate(self, target_type: str) -> URIRef:
        return self._namespace[f"relationship{target_type.capitalize()}"]


class LookupRelationshipSourceTarget(BaseTransformerStandardised):
    """When relationships are extracted, the source and target are extracted as literals. This transformers
    lookup the externalID of the source and target and replaces the literals with the URIRef of the entity.
    """

    description = "Lookup relationships source and target externalId"
    _use_only_once: bool = True
    _need_changes = frozenset({extractors.RelationshipsExtractor.__name__})

    _lookup_entity_query = """SELECT ?entity
    WHERE {{
        ?entity a <{entity_type}> .
        ?entity <{namespace}externalId> "{external_id}" .
    }}"""

    def __init__(self, namespace: Namespace = CLASSIC_CDF_NAMESPACE, type_prefix: str | None = None) -> None:
        self._namespace = namespace
        self._type_prefix = type_prefix
        self._lookup_entity: Callable[[URIRef, str], URIRef] | None = None

    def _count_query(self) -> str:
        return f"""SELECT (COUNT(?instance) AS ?instanceCount)
WHERE {{
  ?instance a <{self._namespace}ClassicRelationship> .
}}"""

    def _iterate_query(self) -> str:
        return f"""SELECT ?instance ?source ?sourceType ?target ?targetType
        WHERE {{
          ?instance a <{self._namespace}ClassicRelationship> .
          ?instance <{self._namespace}sourceExternalId> ?source .
          ?instance <{self._namespace}targetExternalId> ?target .
          ?instance <{self._namespace}sourceType> ?sourceType .
          ?instance <{self._namespace}targetType> ?targetType
        }}"""

    def _iterator(self, graph: Graph) -> Iterator:
        self._lookup_entity = self.create_lookup_entity_with_external_id(graph, self._namespace, self._type_prefix)
        yield from graph.query(self._iterate_query())

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        output = RowTransformationOutput()
        instance, source, source_type, target, target_type = cast(
            tuple[URIRef, Literal, URIRef, Literal, URIRef], query_result_row
        )
        if self._lookup_entity is None:
            raise NeatValueError(f"{type(self)}: .operation() called before .transform()")
        try:
            source_id = self._lookup_entity(source_type, source.toPython())
        except ValueError:
            warnings.warn(ResourceNotFoundWarning(source, "class", str(instance), "class"), stacklevel=2)
            return output

        try:
            target_id = self._lookup_entity(target_type, target.toPython())
        except ValueError:
            warnings.warn(ResourceNotFoundWarning(target, "class", str(instance), "class"), stacklevel=2)
            return output

        output.remove_triples.add((instance, self._namespace.sourceExternalId, source))
        output.remove_triples.add((instance, self._namespace.targetExternalId, target))
        output.add_triples.add((instance, self._namespace.sourceExternalId, source_id))
        output.add_triples.add((instance, self._namespace.targetExternalId, target_id))
        output.instances_modified_count += 1
        return output

    @staticmethod
    def create_lookup_entity_with_external_id(
        graph: Graph, namespace: Namespace, type_prefix: str | None
    ) -> Callable[[URIRef, str], URIRef]:
        @lru_cache(maxsize=10_000)
        def lookup_entity_with_external_id(entity_type: URIRef, external_id: str) -> URIRef:
            if type_prefix:
                entity_type = namespace[type_prefix + remove_namespace_from_uri(entity_type)]

            query = LookupRelationshipSourceTarget._lookup_entity_query.format(
                namespace=namespace, entity_type=entity_type, external_id=external_id
            )
            result = list(graph.query(query))
            if len(result) == 1:
                return cast(URIRef, result[0][0])  # type: ignore[index]
            raise ValueError(f"Could not find entity with external_id {external_id} and type {entity_type}")

        return lookup_entity_with_external_id



from abc import abstractmethod
from collections.abc import Iterable
from typing import TYPE_CHECKING

from rdflib import URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._rules.models import InformationRules
from cognite.neat._shared import Triple
from cognite.neat._utils.auxiliary import class_html_doc

if TYPE_CHECKING:
    from cognite.neat._store._provenance import Agent as ProvenanceAgent


class BaseExtractor:
    """This is the base class for all extractors. It defines the interface that
    extractors must implement.
    """

    def _get_activity_names(self) -> list[str]:
        """Returns the name of the activities that the extractor performs,
        i.e., the actions that it performs when you call extract().."""
        # This method can be overridden by subclasses that runs multiple extractors
        # for example the ClassicGraphExtractor
        return [type(self).__name__]

    @abstractmethod
    def extract(self) -> Iterable[Triple]:
        raise NotImplementedError()

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls)


class KnowledgeGraphExtractor(BaseExtractor):
    """A knowledge graph extractor extracts triples with a schema"""

    @abstractmethod
    def get_information_rules(self) -> InformationRules:
        """Returns the information rules that the extractor uses."""
        raise NotImplementedError()

    @property
    def description(self) -> str:
        return self.__doc__.strip().split("\n")[0] if self.__doc__ else "Missing"

    @property
    def source_uri(self) -> URIRef:
        raise NotImplementedError

    @property
    def agent(self) -> "ProvenanceAgent":
        """Provenance agent for the importer."""
        from cognite.neat._store._provenance import Agent as ProvenanceAgent

        return ProvenanceAgent(id_=DEFAULT_NAMESPACE[f"agent/{type(self).__name__}"])



import sys
import warnings
from collections.abc import Iterable
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, cast, overload
from zipfile import ZipExtFile

import pandas as pd
from pandas import Index
from rdflib import Dataset, Graph, Namespace, URIRef
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore

from cognite.neat._graph._shared import quad_formats, rdflib_to_oxi_type
from cognite.neat._graph.extractors import RdfFileExtractor, TripleExtractors
from cognite.neat._graph.queries import Queries
from cognite.neat._graph.transformers import Transformers
from cognite.neat._issues import IssueList, catch_issues
from cognite.neat._issues.errors import OxigraphStorageLockedError
from cognite.neat._shared import InstanceType, Triple
from cognite.neat._utils.auxiliary import local_import
from cognite.neat._utils.rdf_ import add_triples_in_batch, remove_namespace_from_uri

from ._provenance import Change, Entity, Provenance

if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self


class NeatGraphStore:
    """NeatGraphStore is a class that stores the graph and provides methods to read/write data it contains


    Args:
        graph : Instance of rdflib.Graph class for graph storage
        rules:

    !!! note "Dataset"
        The store leverages a RDF dataset which is defined as a collection of RDF graphs
        where all but one are named graphs associated with URIRef (the graph name),
        and the unnamed default graph which is in context of rdflib library has an
        identifier URIRef('urn:x-rdflib:default').
    """

    rdf_store_type: str

    def __init__(
        self,
        dataset: Dataset,
        default_named_graph: URIRef | None = None,
    ):
        _start = datetime.now(timezone.utc)
        self.dataset = dataset
        self.provenance = Provenance[Entity](
            [
                Change.record(
                    activity=f"{type(self).__name__}.__init__",
                    start=_start,
                    end=datetime.now(timezone.utc),
                    description=f"Initialize graph store as {type(self.dataset.store).__name__}",
                )
            ]
        )

        self.default_named_graph = default_named_graph or DATASET_DEFAULT_GRAPH_ID
        self.queries = Queries(self.dataset, self.default_named_graph)

    def graph(self, named_graph: URIRef | None = None) -> Graph:
        """Get named graph from the dataset to query over"""
        return self.dataset.graph(named_graph or self.default_named_graph)

    @property
    def type_(self) -> str:
        "Return type of the graph store"
        return type(self.dataset.store).__name__

    # no destination
    @overload
    def serialize(self, filepath: None = None) -> str: ...

    # with destination
    @overload
    def serialize(self, filepath: Path) -> None: ...

    def serialize(self, filepath: Path | None = None) -> None | str:
        """Serialize the graph store to a file.

        Args:
            filepath: File path to serialize the graph store to

        Returns:
            Serialized graph store

        !!! note "Trig Format"
            Notice that instead of turtle format we are using trig format for serialization.
            This is because trig format is a superset of turtle format and it allows us to
            serialize named graphs as well. Allowing serialization of one or more named graphs
            including the default graph.
        """
        if filepath:
            self.dataset.serialize(
                filepath,
                format="ox-trig" if self.type_ == "OxigraphStore" else "trig",
            )
            return None
        else:
            return self.dataset.serialize(format="ox-trig" if self.type_ == "OxigraphStore" else "trig")

    def _upsert_prefixes(self, prefixes: dict[str, Namespace], named_graph: URIRef) -> None:
        """Adds prefixes to the graph store."""
        _start = datetime.now(timezone.utc)
        for prefix, namespace in prefixes.items():
            self.graph(named_graph).bind(prefix, namespace)

        self.provenance.append(
            Change.record(
                activity=f"{type(self).__name__}._upsert_prefixes",
                start=_start,
                end=datetime.now(timezone.utc),
                description="Upsert prefixes to the name graph {named_graph}",
            )
        )

    @classmethod
    def from_memory_store(cls) -> "Self":
        return cls(Dataset())

    @classmethod
    def from_sparql_store(
        cls,
        query_endpoint: str | None = None,
        update_endpoint: str | None = None,
        returnFormat: str = "csv",
    ) -> "Self":
        store = SPARQLUpdateStore(
            query_endpoint=query_endpoint,
            update_endpoint=update_endpoint,
            returnFormat=returnFormat,
            context_aware=False,
            postAsEncoded=False,
            autocommit=False,
        )
        graph = Dataset(store=store)
        return cls(graph)

    @classmethod
    def from_oxi_remote_store(
        cls,
        remote_url: str,
        autocommit: bool = False,
    ) -> "Self":
        """Creates a NeatGraphStore from a remote Oxigraph store SPARQL endpoint."""

        return cls(
            dataset=Dataset(
                store=SPARQLUpdateStore(
                    query_endpoint=f"{remote_url}/query", update_endpoint=f"{remote_url}/query", autocommit=autocommit
                ),
                default_union=True,
            )
        )

    @classmethod
    def from_oxi_local_store(cls, storage_dir: Path | None = None) -> "Self":
        """Creates a NeatGraphStore from an Oxigraph store."""
        local_import("pyoxigraph", "oxi")
        local_import("oxrdflib", "oxi")
        import oxrdflib
        import pyoxigraph

        try:
            oxi_store = pyoxigraph.Store(path=str(storage_dir) if storage_dir else None)
        except OSError as e:
            if "lock" in str(e):
                raise OxigraphStorageLockedError(filepath=cast(Path, storage_dir)) from e
            raise e

        return cls(
            dataset=Dataset(
                store=oxrdflib.OxigraphStore(store=oxi_store),
            )
        )

    def write(self, extractor: TripleExtractors, named_graph: URIRef | None = None) -> IssueList:
        last_change: Change | None = None
        named_graph = named_graph or self.default_named_graph
        with catch_issues() as issue_list:
            _start = datetime.now(timezone.utc)
            success = True

            if isinstance(extractor, RdfFileExtractor) and not extractor.issue_list.has_errors:
                self._parse_file(
                    named_graph,
                    extractor.filepath,
                    cast(str, extractor.format),
                    extractor.base_uri,
                )
                if isinstance(extractor.filepath, ZipExtFile):
                    extractor.filepath.close()

            elif isinstance(extractor, RdfFileExtractor):
                success = False
                issue_text = "\n".join([issue.as_message() for issue in extractor.issue_list])
                warnings.warn(
                    (
                        f"Cannot write to named graph {named_graph} with "
                        f"{type(extractor).__name__}, errors found in file:\n{issue_text}"
                    ),
                    stacklevel=2,
                )
            else:
                self._add_triples(extractor.extract(), named_graph=named_graph)

            if success:
                _end = datetime.now(timezone.utc)
                # Need to do the hasattr in case the extractor comes from NeatEngine.
                activities = (
                    extractor._get_activity_names()
                    if hasattr(extractor, "_get_activity_names")
                    else [type(extractor).__name__]
                )
                for activity in activities:
                    last_change = Change.record(
                        activity=activity,
                        start=_start,
                        end=_end,
                        description=f"Extracted triples to named graph {named_graph} using {type(extractor).__name__}",
                    )
                    self.provenance.append(last_change)
        if last_change:
            last_change.target_entity.issues.extend(issue_list)
        return issue_list

    def read(
        self,
        class_uri: URIRef,
        named_graph: URIRef | None = None,
        property_renaming_config: dict[URIRef, str] | None = None,
        remove_uri_namespace: bool = True,
    ) -> Iterable[tuple[URIRef, dict[str | InstanceType, list[Any]]]]:
        named_graph = named_graph or self.default_named_graph

        instance_ids = self.queries.list_instances_ids(class_uri, named_graph=named_graph)

        for instance_id in instance_ids:
            if res := self.queries.describe(
                instance_id=instance_id,
                instance_type=class_uri,
                property_renaming_config=property_renaming_config,
                remove_uri_namespace=remove_uri_namespace,
            ):
                yield res

    def _parse_file(
        self,
        named_graph: URIRef,
        filepath: Path | ZipExtFile,
        format: str = "turtle",
        base_uri: URIRef | None = None,
    ) -> None:
        """Imports graph data from file.

        Args:
            named_graph : URIRef of the named graph to store the data in
            filepath : File path to file containing graph data, by default None
            format : rdflib format file containing RDF graph, by default "turtle"
            base_uri : base URI to add to graph in case of relative URIs, by default None

        !!! note "Oxigraph store"
            By default we are using non-transactional mode for parsing RDF files.
            This gives us a significant performance boost when importing large RDF files.
            Underhood of rdflib we are triggering oxrdflib plugin which in respect
            calls `bulk_load` method from oxigraph store. See more at:
            https://pyoxigraph.readthedocs.io/en/stable/store.html#pyoxigraph.Store.bulk_load
        """

        # Oxigraph store, do not want to type hint this as it is an optional dependency
        if self.type_ == "OxigraphStore":
            local_import("pyoxigraph", "oxi")

            if format in quad_formats():
                self.dataset.parse(
                    filepath,  # type: ignore[arg-type]
                    format=rdflib_to_oxi_type(format),
                    transactional=False,
                    publicID=base_uri,
                )
            else:
                self.graph(named_graph).parse(
                    filepath,  # type: ignore[arg-type]
                    format=rdflib_to_oxi_type(format),
                    transactional=False,
                    publicID=base_uri,
                )
            self.dataset.store._store.optimize()  # type: ignore[attr-defined]

        # All other stores
        else:
            if format in quad_formats():
                self.dataset.parse(filepath, publicID=base_uri, format=format)  # type: ignore[arg-type]
            else:
                self.graph(named_graph).parse(filepath, publicID=base_uri, format=format)  # type: ignore[arg-type]

    def _add_triples(
        self,
        triples: Iterable[Triple],
        named_graph: URIRef,
        batch_size: int = 10_000,
    ) -> None:
        """Adds triples to the graph store in batches.

        Args:
            triples: list of triples to be added to the graph store
            batch_size: Batch size of triples per commit, by default 10_000
            verbose: Verbose mode, by default False
        """
        add_triples_in_batch(self.graph(named_graph), triples, batch_size)

    def transform(self, transformer: Transformers, named_graph: URIRef | None = None) -> None:
        """Transforms the graph store using a transformer."""

        named_graph = named_graph or self.default_named_graph
        if named_graph in self.named_graphs:
            missing_changes = [
                change for change in transformer._need_changes if not self.provenance.activity_took_place(change)
            ]
            if self.provenance.activity_took_place(type(transformer).__name__) and transformer._use_only_once:
                warnings.warn(
                    f"Cannot transform graph store with {type(transformer).__name__}, already applied",
                    stacklevel=2,
                )
            elif missing_changes:
                warnings.warn(
                    (
                        f"Cannot transform graph store with {type(transformer).__name__}, "
                        f"missing one or more required changes [{', '.join(missing_changes)}]"
                    ),
                    stacklevel=2,
                )

            else:
                _start = datetime.now(timezone.utc)
                transformer.transform(self.graph(named_graph))
                self.provenance.append(
                    Change.record(
                        activity=f"{type(transformer).__name__}",
                        start=_start,
                        end=datetime.now(timezone.utc),
                        description=transformer.description,
                    )
                )

        else:
            warnings.warn(
                f"Named graph {named_graph} not found in graph store, cannot transform",
                stacklevel=2,
            )

    @property
    def summary(self) -> dict[URIRef, pd.DataFrame]:
        return {
            named_graph: pd.DataFrame(
                self.queries.summarize_instances(named_graph),
                columns=["Type", "Occurrence"],
            )
            for named_graph in self.named_graphs
        }

    @property
    def multi_type_instances(self) -> dict[URIRef, dict[str, list[str]]]:
        return {named_graph: self.queries.multi_type_instances(named_graph) for named_graph in self.named_graphs}

    def _repr_html_(self) -> str:
        provenance = self.provenance._repr_html_()
        summary: dict[URIRef, pd.DataFrame] = self.summary

        def _short_name_of_graph(named_graph: URIRef) -> str:
            return "default" if named_graph == self.default_named_graph else remove_namespace_from_uri(named_graph)

        if not summary:
            summary_text = "<br /><strong>Graph is empty</strong><br />"
        else:
            all_types = set().union(
                *[set(sub_summary.Type) for sub_summary in summary.values() if not sub_summary.empty]
            )

            summary_text = (
                "<br /><strong>Overview</strong>:"  # type: ignore
                f"<ul><li>{len(summary)} named graphs</strong></li>"
                f"<li>Total of {len(all_types)} unique types</strong></li>"
            )

            for named_graph, table in summary.items():
                summary_text += (
                    f"<li>{sum(table['Occurrence'])} instances in {_short_name_of_graph(named_graph)}"
                    " graph</strong></li>"
                )

            summary_text += "</ul>"
            for named_graph, table in summary.items():
                summary_text += (
                    f"<br /><strong>{_short_name_of_graph(named_graph)} graph</strong>:"
                    f"{cast(pd.DataFrame, self._shorten_summary(table))._repr_html_()}"  # type: ignore[operator]
                )

        for named_graph, multi_value_instances in self.multi_type_instances.items():
            if multi_value_instances:
                summary_text += (
                    f"<br><strong>Multi value instances detected in {_short_name_of_graph(named_graph)}"
                    "graph! Loading could have issues!</strong></br>"
                )

        return f"{summary_text}{provenance}"

    def _shorten_summary(self, summary: pd.DataFrame) -> pd.DataFrame:
        """Shorten summary to top 5 types by occurrence."""
        top_5_rows = summary.head(5)

        indexes = [
            *top_5_rows.index.tolist(),
        ]
        data = [
            top_5_rows,
        ]
        if len(summary) > 6:
            last_row = summary.tail(1)
            indexes += [
                "...",
                *last_row.index.tolist(),
            ]
            data.extend([pd.DataFrame([["..."] * summary.shape[1]], columns=summary.columns), last_row])

        shorter_summary = pd.concat(
            data,
            ignore_index=True,
        )
        shorter_summary.index = cast(Index, indexes)

        return shorter_summary

    @property
    def named_graphs(self) -> list[URIRef]:
        return [cast(URIRef, context.identifier) for context in self.dataset.contexts()]

    @property
    def empty(self) -> bool:
        """Cheap way to check if the graph store is empty."""
        return not self.queries.has_data()



import functools
import warnings
from collections.abc import Callable
from typing import Any

from cognite.neat._alpha import AlphaWarning
from cognite.neat._issues.errors import CDFMissingClientError, NeatImportError
from cognite.neat._issues.errors._external import OxigraphStorageLockedError
from cognite.neat._issues.errors._general import NeatValueError

from ._collector import _COLLECTOR

try:
    from rich import print
    from rich.markup import escape

    _ERROR_PREFIX = "[bold red][ERROR][/bold red]"
    _WARNING_PREFIX = "[bold bright_magenta][WARNING][/bold bright_magenta]"
except ImportError:
    _ERROR_PREFIX = "[ERROR]"
    _WARNING_PREFIX = "[WARNING]"

    def escape(x: Any, *_: Any, **__: Any) -> Any:  # type: ignore[misc]
        return x


class NeatSessionError(Exception):
    """Base class for all exceptions raised by the NeatSession class."""

    ...


def _session_method_wrapper(func: Callable, cls_name: str):
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any):
        _COLLECTOR.track_session_command(f"{cls_name}.{func.__name__}", *args, **kwargs)
        try:
            with warnings.catch_warnings(record=True) as w:
                result = func(*args, **kwargs)
                for warning in w:
                    if isinstance(warning.message, AlphaWarning):
                        print(f"{_WARNING_PREFIX} {warning.message}")

            return result
        except NeatSessionError as e:
            action = _get_action()
            print(f"{_ERROR_PREFIX} Cannot {action}: {e}")
        except (
            CDFMissingClientError,
            NeatImportError,
            NeatValueError,
            OxigraphStorageLockedError,
        ) as e:
            print(f"{_ERROR_PREFIX} {escape(e.as_message())}")
        except ModuleNotFoundError as e:
            if e.name == "neatengine":
                action = _get_action()
                print(f"{_ERROR_PREFIX} The functionality {action} requires the NeatEngine.")
            else:
                raise e

    def _get_action():
        action = func.__name__
        if action == "__call__":
            action = func.__qualname__.removesuffix(".__call__").removesuffix("API")
        return action

    return wrapper


def session_class_wrapper(cls: type):
    """This decorator wraps all methods of a class.

    It should be used with all composition classes used with the NeatSession class.

    It does the following:
        * Intercepts all NeatSession exceptions and prints them in a user-friendly way.
        * Collects user metrics.

    Args:
        cls: NeatSession composition class

    Returns:
        cls: NeatSession composition class with all methods wrapped
    """
    to_check = [cls]
    while to_check:
        cls = to_check.pop()
        for attr_name in dir(cls):
            if attr_name.startswith("_") and not attr_name == "__call__":
                continue
            attr = getattr(cls, attr_name)
            if callable(attr):
                setattr(cls, attr_name, _session_method_wrapper(attr, cls.__name__))
            elif isinstance(attr, type):
                to_check.append(attr)
    return cls



from dataclasses import dataclass
from typing import Generic

from cognite.neat._issues._base import NeatWarning, ResourceType, T_Identifier, T_ReferenceIdentifier


# Name ResourceNeatWarning to avoid conflicts with the built-in ResourceWarning
@dataclass(unsafe_hash=True)
class ResourceNeatWarning(NeatWarning, Generic[T_Identifier]):
    """Base class for resource warnings {resource_type} with identifier {identifier}"""

    identifier: T_Identifier
    resource_type: ResourceType


@dataclass(unsafe_hash=True)
class ResourceRegexViolationWarning(ResourceNeatWarning):
    """The {resource_type} with identifier {identifier} in the {location} is violating
    the CDF regex {regex}. This will lead to errors when converting to DMS data model.
    """

    fix = (
        "Either export the data model and make the necessary changes manually"
        " or run fix.data_model.cdf_compliant_external_ids."
    )

    location: str
    regex: str


@dataclass(unsafe_hash=True)
class ResourceNotFoundWarning(ResourceNeatWarning, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier {identifier} referred by {referred_type} {referred_by} does not exist.
    This will be ignored."""

    fix = "Create the {resource_type}"

    referred_by: T_ReferenceIdentifier
    referred_type: str


@dataclass(unsafe_hash=True)
class ResourceUnknownWarning(ResourceNeatWarning, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier {identifier} referred by {referred_type} {referred_by} is unknown.
    Will continue, but the model is incomplete."""

    referred_by: T_ReferenceIdentifier
    referred_type: str

    fix = "You can maybe retrieve the resource from the CDF."


@dataclass(unsafe_hash=True)
class ResourceNotDefinedWarning(ResourceNeatWarning, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} {identifier} is not defined in the {location}"""

    extra = "{column_name} {row_number} in {sheet_name}"
    fix = "Define the {resource_type} {identifier} in {location}."

    location: str
    column_name: str | None = None
    row_number: int | None = None
    sheet_name: str | None = None


@dataclass(unsafe_hash=True)
class ResourceRedefinedWarning(ResourceNeatWarning, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} {identifier} feature {feature} is being redefine from {current_value} to {new_value}.
    This will be ignored."""

    fix = "Avoid redefinition {resource_type} features"
    feature: str
    current_value: str
    new_value: str


@dataclass(unsafe_hash=True)
class ResourcesDuplicatedWarning(NeatWarning, Generic[T_Identifier]):
    """Duplicated {resource_type} with identifiers {resources} were found. {default_action}"""

    fix = "Remove the duplicate resources"

    resources: frozenset[T_Identifier]
    resource_type: ResourceType
    default_action: str


@dataclass(unsafe_hash=True)
class ResourceRetrievalWarning(NeatWarning, Generic[T_Identifier]):
    """Failed to retrieve {resource_type} with identifier(s) {resources}. Continuing without
    these resources."""

    extra = "The error was: {error}"

    fix = "Check the error and fix accordingly."

    resources: frozenset[T_Identifier]
    resource_type: ResourceType
    error: str | None = None


class ResourceTypeNotSupportedWarning(ResourceNeatWarning[T_Identifier]):
    """The {resource_type} with identifier {identifier} is not supported. This will be ignored."""

    resource_type: str  # type: ignore[assignment]



from logging import getLogger

from cognite.neat._issues import NeatError, NeatIssue, NeatWarning

from .base import Tracker


class LogTracker(Tracker):
    def __init__(self, name: str, units: list[str], unit_type: str) -> None:
        super().__init__(name, units, unit_type)
        self._logger = getLogger(__name__)
        self._total_units = len(units)
        self._count = 1
        self._logger.info(f"Staring {self.name} and will process {len(units)} {unit_type}.")

    def start(self, unit: str) -> None:
        self._logger.info(f"Starting {unit} {self._count}/{self._total_units}.")
        self._count += 1

    def finish(self, unit: str) -> None:
        self._logger.info(f"Finished {unit}.")

    def _issue(self, issue: NeatIssue) -> None:
        if isinstance(issue, NeatWarning):
            self._logger.warning(issue)
        elif isinstance(issue, NeatError):
            self._logger.error(issue)



from dataclasses import dataclass, field
from typing import Any, Literal, cast, overload

import pandas as pd
from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet

from cognite.neat._rules._constants import get_internal_properties


@dataclass
class SpreadsheetRead:
    """This class is used to store information about the source spreadsheet.

    It is used to adjust row numbers to account for header rows and empty rows
    such that the error/warning messages are accurate.
    """

    header_row: int = 1
    empty_rows: list[int] = field(default_factory=list)
    skipped_rows: list[int] = field(default_factory=list)
    is_one_indexed: bool = True

    def __post_init__(self):
        self.empty_rows = sorted(self.empty_rows)

    def adjusted_row_number(self, row_no: int) -> int:
        output = row_no
        for empty_row in self.empty_rows:
            if empty_row <= output:
                output += 1
            else:
                break

        for skipped_rows in self.skipped_rows:
            if skipped_rows <= output:
                output += 1
            else:
                break

        return output + self.header_row + (1 if self.is_one_indexed else 0)


@overload
def read_individual_sheet(
    excel_file: pd.ExcelFile,
    sheet_name: str,
    return_read_info: Literal[True],
    expected_headers: list[str] | None = None,
) -> tuple[list[dict], SpreadsheetRead]: ...


@overload
def read_individual_sheet(
    excel_file: pd.ExcelFile,
    sheet_name: str,
    return_read_info: Literal[False] = False,
    expected_headers: list[str] | None = None,
) -> list[dict]: ...


def read_individual_sheet(
    excel_file: pd.ExcelFile,
    sheet_name: str,
    return_read_info: bool = False,
    expected_headers: list[str] | None = None,
) -> tuple[list[dict], SpreadsheetRead] | list[dict]:
    if expected_headers:
        target_row = _get_row_number(cast(Worksheet, load_workbook(excel_file)[sheet_name]), expected_headers)
        skiprows = target_row - 1 if target_row is not None else 0
    else:
        skiprows = 0

    raw = pd.read_excel(excel_file, sheet_name, skiprows=skiprows)
    is_na = raw.isnull().all(axis=1)
    skip_rows = _find_rows_to_skip(raw)
    empty_rows = is_na[is_na].index.tolist()

    if skip_rows:
        raw = raw.drop(skip_rows)

    raw.dropna(axis=0, how="all", inplace=True)

    if "Value Type" in raw.columns:
        # Special handling for Value Type column, #N/A is treated specially by NEAT it means Unknown
        raw["Value Type"] = raw["Value Type"].replace(float("nan"), "#N/A")

    output = raw.replace(float("nan"), None).to_dict(orient="records")
    if return_read_info:
        # If no rows are skipped, row 1 is the header row.
        return output, SpreadsheetRead(
            header_row=skiprows + 1,
            empty_rows=empty_rows,
            is_one_indexed=True,
            skipped_rows=skip_rows,
        )
    return output


def _find_rows_to_skip(
    df: pd.DataFrame,
) -> list:
    """Find rows which are having all values as None except for internal properties."""
    rows_to_skip = []

    internal_cols = {val.lower() for val in get_internal_properties()}
    for i, row in df.iterrows():
        user_cols_state = []
        internal_cols_state = []
        for col in df.columns:
            if col.lower() not in internal_cols:
                user_cols_state.append(row[col] == "#N/A" or row[col].__str__().lower() in ["none", "nan"])
            else:
                internal_cols_state.append(row[col] is not None)

        if all(user_cols_state) and any(internal_cols_state):
            rows_to_skip.append(i)

    return rows_to_skip


def _get_row_number(sheet: Worksheet, values_to_find: list[str]) -> int | None:
    for row_number, row in enumerate(sheet.iter_rows(values_only=True), start=1):
        if any(value in row for value in values_to_find):
            return row_number
    return None


def find_column_with_value(sheet: Worksheet, value: Any) -> str | None:
    for row in sheet.iter_rows():
        for cell in row:
            if cell.value and isinstance(cell.value, str) and cell.value.lower() == value.lower():
                return cell.column_letter  # type: ignore

    return None



import warnings
from collections.abc import Callable
from typing import Annotated, Any, TypeAlias, TypeVar

import rdflib
from pydantic import (
    AfterValidator,
    BeforeValidator,
    HttpUrl,
    StringConstraints,
    TypeAdapter,
    ValidationInfo,
    WrapValidator,
)
from pydantic.functional_serializers import PlainSerializer

from cognite.neat._issues.errors import RegexViolationError
from cognite.neat._issues.warnings import RegexViolationWarning
from cognite.neat._rules._constants import (
    DATA_MODEL_COMPLIANCE_REGEX,
    PATTERNS,
    PREFIX_COMPLIANCE_REGEX,
    VERSION_COMPLIANCE_REGEX,
    EntityTypes,
)
from cognite.neat._rules.models.entities._multi_value import MultiValueTypeInfo
from cognite.neat._rules.models.entities._single_value import (
    ClassEntity,
    ContainerEntity,
    ViewEntity,
)

Entities: TypeAlias = ClassEntity | ViewEntity | ContainerEntity
T_Entities = TypeVar("T_Entities", bound=Entities)


def _custom_error(exc_factory: Callable[[str | None, Exception], Any]) -> Any:
    def _validator(value: Any, next_: Any, ctx: ValidationInfo) -> Any:
        try:
            return next_(value, ctx)
        except ValueError as e:
            if value is None:
                raise e
            raise exc_factory(ctx.field_name, value) from None

    return WrapValidator(_validator)


StrOrListType = Annotated[
    str | list[str],
    BeforeValidator(lambda value: value.replace(", ", ",").split(",") if isinstance(value, str) and value else value),
    PlainSerializer(
        lambda value: ",".join([entry for entry in value if entry]) if isinstance(value, list) else value,
        return_type=str,
        when_used="unless-none",
    ),
]

StrListType = Annotated[
    list[str],
    BeforeValidator(lambda value: [entry.strip() for entry in value.split(",")] if isinstance(value, str) else value),
    PlainSerializer(
        lambda value: ",".join([entry for entry in value if entry]), return_type=str, when_used="unless-none"
    ),
]

NamespaceType = Annotated[
    rdflib.Namespace,
    BeforeValidator(
        lambda value: (
            rdflib.Namespace(TypeAdapter(HttpUrl).validate_python(value))
            if value.endswith("#") or value.endswith("/")
            else rdflib.Namespace(TypeAdapter(HttpUrl).validate_python(f"{value}/"))
        )
    ),
]

URIRefType = Annotated[
    rdflib.URIRef,
    BeforeValidator(lambda value: rdflib.URIRef(value)),
    PlainSerializer(
        lambda value: str(value),
        return_type=str,
        when_used="unless-none",
    ),
]

PrefixType = Annotated[
    str,
    StringConstraints(pattern=PREFIX_COMPLIANCE_REGEX),
    _custom_error(
        lambda _, value: RegexViolationError(value, PREFIX_COMPLIANCE_REGEX, "prefix entry in metadata"),
    ),
]

DataModelExternalIdType = Annotated[
    str,
    StringConstraints(pattern=DATA_MODEL_COMPLIANCE_REGEX),
    _custom_error(
        lambda _, value: RegexViolationError(value, DATA_MODEL_COMPLIANCE_REGEX, "external ID entry in metadata")
    ),
]


VersionType = Annotated[
    str,
    BeforeValidator(str),
    StringConstraints(pattern=VERSION_COMPLIANCE_REGEX),
    _custom_error(lambda _, value: RegexViolationError(value, VERSION_COMPLIANCE_REGEX, "version entry in metadata")),
]


def _external_id_validation_factory(entity_type: EntityTypes, location: str):
    def _external_id_validation(value: str) -> str:
        compiled_regex = PATTERNS.entity_pattern(entity_type)
        if not compiled_regex.match(value):
            raise RegexViolationError(value, compiled_regex.pattern, location)
        if PATTERNS.more_than_one_alphanumeric.search(value):
            warnings.warn(
                RegexViolationWarning(
                    value,
                    compiled_regex.pattern,
                    "property",
                    "MoreThanOneNonAlphanumeric",
                ),
                stacklevel=2,
            )
        return value

    return _external_id_validation


SpaceType = Annotated[
    str,
    AfterValidator(_external_id_validation_factory(EntityTypes.space, "")),
]

InformationPropertyType = Annotated[
    str,
    AfterValidator(_external_id_validation_factory(EntityTypes.information_property, "Property column in properties")),
]
DmsPropertyType = Annotated[
    str,
    AfterValidator(_external_id_validation_factory(EntityTypes.dms_property, "Property column in properties")),
]


def _entity_validation(value: Entities, location: str) -> Entities:
    suffix_regex = PATTERNS.entity_pattern(value.type_)
    if not suffix_regex.match(value.suffix):
        raise RegexViolationError(str(value), suffix_regex.pattern, location)
    return value


ClassEntityType = Annotated[ClassEntity, AfterValidator(lambda v: _entity_validation(v, "the Class column"))]
ViewEntityType = Annotated[ViewEntity, AfterValidator(lambda v: _entity_validation(v, "the View column"))]
ContainerEntityType = Annotated[
    ContainerEntity, AfterValidator(lambda v: _entity_validation(v, "the Container column"))
]


def _multi_value_type_validation(value: MultiValueTypeInfo, location: str) -> MultiValueTypeInfo:
    for type_ in value.types:
        if isinstance(type_, ClassEntity):
            _entity_validation(type_, location)
    return value


MultiValueTypeType = Annotated[
    MultiValueTypeInfo, AfterValidator(lambda v: _multi_value_type_validation(v, "the Value Type column"))
]



import zipfile
from collections.abc import Iterable
from pathlib import Path
from typing import cast, get_args

from rdflib import URIRef
from rdflib.util import guess_format

from cognite.neat._constants import DEFAULT_BASE_URI
from cognite.neat._graph._shared import RDFTypes
from cognite.neat._graph.extractors._base import BaseExtractor
from cognite.neat._issues._base import IssueList
from cognite.neat._issues.errors import FileNotFoundNeatError, FileTypeUnexpectedError
from cognite.neat._issues.errors._general import NeatValueError
from cognite.neat._shared import Triple


class RdfFileExtractor(BaseExtractor):
    """Extract data from RDF files into Neat.

    Args:
        filepath (Path): The path to the RDF file.
        mime_type (MIMETypes, optional): The MIME type of the RDF file. Defaults to "application/rdf+xml".
        base_uri (URIRef, optional): The base URI to use. Defaults to None.
    """

    def __init__(
        self,
        filepath: Path | zipfile.ZipExtFile,
        base_uri: URIRef = DEFAULT_BASE_URI,
        issue_list: IssueList | None = None,
    ):
        self.issue_list = issue_list or IssueList(title=f"{filepath.name}")
        self.base_uri = base_uri
        self.filepath = filepath

        self.format = guess_format(str(self.filepath) if isinstance(self.filepath, Path) else self.filepath.name)

        if isinstance(self.filepath, Path) and not self.filepath.exists():
            self.issue_list.append(FileNotFoundNeatError(self.filepath))

        if not self.format:
            self.issue_list.append(
                FileTypeUnexpectedError(
                    (self.filepath if isinstance(self.filepath, Path) else Path(self.filepath.name)),
                    frozenset(get_args(RDFTypes)),
                )
            )

    def extract(self) -> Iterable[Triple]:
        raise NotImplementedError()

    @classmethod
    def from_zip(
        cls,
        filepath: Path,
        filename: str = "neat-session/instances/instances.trig",
        base_uri: URIRef = DEFAULT_BASE_URI,
        issue_list: IssueList | None = None,
    ):
        if not filepath.exists():
            raise FileNotFoundNeatError(filepath)
        if filepath.suffix not in {".zip"}:
            raise NeatValueError("Expected a zip file, got {filepath.suffix}")

        with zipfile.ZipFile(filepath, "r") as zip_ref:
            for file_info in zip_ref.infolist():
                if file_info.filename == filename:
                    # We need to open the file in the zip file, and close it upon
                    # triple extraction ...

                    print(file_info)
                    file = zip_ref.open(file_info)
                    return cls(cast(zipfile.ZipExtFile, file), base_uri, issue_list)

        raise NeatValueError(f"Cannot extract {filename} from zip file {filepath}")



import itertools
import warnings
from collections import defaultdict
from collections.abc import Hashable, ItemsView, Iterator, KeysView, MutableMapping, Set, ValuesView
from dataclasses import dataclass, field
from graphlib import TopologicalSorter
from typing import Any, Literal, TypeVar, overload

import networkx as nx
import pandas as pd
from cognite.client import data_modeling as dm
from rdflib import URIRef

from cognite.neat._issues.errors import NeatValueError
from cognite.neat._issues.warnings import NeatValueWarning
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.models.dms import DMSProperty
from cognite.neat._rules.models.dms._rules import DMSView
from cognite.neat._rules.models.entities import ClassEntity, MultiValueTypeInfo, ViewEntity
from cognite.neat._rules.models.entities._single_value import (
    UnknownEntity,
)
from cognite.neat._rules.models.information import InformationClass, InformationProperty

T_Hashable = TypeVar("T_Hashable", bound=Hashable)


@dataclass(frozen=True)
class Linkage:
    source_class: ClassEntity
    connecting_property: str
    target_class: ClassEntity
    max_occurrence: int | float | None


class LinkageSet(set, Set[Linkage]):
    @property
    def source_class(self) -> set[ClassEntity]:
        return {link.source_class for link in self}

    @property
    def target_class(self) -> set[ClassEntity]:
        return {link.target_class for link in self}

    def get_target_classes_by_source(self) -> dict[ClassEntity, set[ClassEntity]]:
        target_classes_by_source: dict[ClassEntity, set[ClassEntity]] = defaultdict(set)
        for link in self:
            target_classes_by_source[link.source_class].add(link.target_class)
        return target_classes_by_source

    def to_pandas(self) -> pd.DataFrame:
        # Todo: Remove this method
        return pd.DataFrame(
            [
                {
                    "source_class": link.source_class,
                    "connecting_property": link.connecting_property,
                    "target_class": link.target_class,
                    "max_occurrence": link.max_occurrence,
                }
                for link in self
            ]
        )


@dataclass
class ViewQuery:
    view_id: dm.ViewId
    rdf_type: URIRef
    property_renaming_config: dict[URIRef, str] = field(default_factory=dict)


class ViewQueryDict(dict, MutableMapping[dm.ViewId, ViewQuery]):
    # The below methods are included to make better type hints in the IDE
    def __getitem__(self, k: dm.ViewId) -> ViewQuery:
        return super().__getitem__(k)

    def __setitem__(self, k: dm.ViewId, v: ViewQuery) -> None:
        super().__setitem__(k, v)

    def __delitem__(self, k: dm.ViewId) -> None:
        super().__delitem__(k)

    def __iter__(self) -> Iterator[dm.ViewId]:
        return super().__iter__()

    def keys(self) -> KeysView[dm.ViewId]:  # type: ignore[override]
        return super().keys()

    def values(self) -> ValuesView[ViewQuery]:  # type: ignore[override]
        return super().values()

    def items(self) -> ItemsView[dm.ViewId, ViewQuery]:  # type: ignore[override]
        return super().items()

    def get(self, __key: dm.ViewId, __default: Any = ...) -> ViewQuery:
        return super().get(__key, __default)

    def pop(self, __key: dm.ViewId, __default: Any = ...) -> ViewQuery:
        return super().pop(__key, __default)

    def popitem(self) -> tuple[dm.ViewId, ViewQuery]:
        return super().popitem()


class RulesAnalysis:
    def __init__(self, information: InformationRules | None = None, dms: DMSRules | None = None) -> None:
        self._information = information
        self._dms = dms

    @property
    def information(self) -> InformationRules:
        if self._information is None:
            raise NeatValueError("Information rules are required for this analysis")
        return self._information

    @property
    def dms(self) -> DMSRules:
        if self._dms is None:
            raise NeatValueError("DMS rules are required for this analysis")
        return self._dms

    def parents_by_class(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ClassEntity, set[ClassEntity]]:
        """Get a dictionary of classes and their parents.

        Args:
            include_ancestors (bool, optional): Include ancestors of the parents. Defaults to False.
            include_different_space (bool, optional): Include parents from different spaces. Defaults to False.

        Returns:
            dict[ClassEntity, set[ClassEntity]]: Values parents with class as key.
        """
        parents_by_class: dict[ClassEntity, set[ClassEntity]] = {}
        for class_ in self.information.classes:
            parents_by_class[class_.class_] = set()
            for parent in class_.implements or []:
                if include_different_space or parent.prefix == class_.class_.prefix:
                    parents_by_class[class_.class_].add(parent)
                else:
                    warnings.warn(
                        NeatValueWarning(
                            f"Parent class {parent} of class {class_} is not in the same namespace, skipping!"
                        ),
                        stacklevel=2,
                    )
        if include_ancestors:
            self._include_ancestors(parents_by_class)

        return parents_by_class

    @staticmethod
    def _include_ancestors(parents_by_class: dict[T_Hashable, set[T_Hashable]]) -> None:
        # Topological sort to ensure that classes include all ancestors
        for class_entity in list(TopologicalSorter(parents_by_class).static_order()):
            parents_by_class[class_entity] |= {
                grand_parent for parent in parents_by_class[class_entity] for grand_parent in parents_by_class[parent]
            }

    def properties_by_class(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ClassEntity, list[InformationProperty]]:
        """Get a dictionary of classes and their properties.

        Args:
            include_ancestors: Whether to include properties from parent classes.
            include_different_space: Whether to include properties from parent classes in different spaces.

        Returns:
            dict[ClassEntity, list[InformationProperty]]: Values properties with class as key.

        """
        properties_by_classes: dict[ClassEntity, list[InformationProperty]] = defaultdict(list)
        for prop in self.information.properties:
            properties_by_classes[prop.class_].append(prop)

        if include_ancestors:
            parents_by_classes = self.parents_by_class(
                include_ancestors=include_ancestors, include_different_space=include_different_space
            )
            for class_, parents in parents_by_classes.items():
                class_properties = {prop.property_ for prop in properties_by_classes[class_]}
                for parent in parents:
                    for parent_prop in properties_by_classes[parent]:
                        if parent_prop.property_ not in class_properties:
                            child_prop = parent_prop.model_copy(update={"class_": class_})
                            properties_by_classes[class_].append(child_prop)
                            class_properties.add(child_prop.property_)

        return properties_by_classes

    def implements_by_view(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ViewEntity, set[ViewEntity]]:
        """Get a dictionary of views and their implemented views."""
        # This is a duplicate fo the parent_by_class method, but for views
        # The choice to duplicate the code is to avoid generics which will make the code less readable
        implements_by_view: dict[ViewEntity, set[ViewEntity]] = {}
        for view in self.dms.views:
            implements_by_view[view.view] = set()
            for implements in view.implements or []:
                if include_different_space or implements.space == view.view.space:
                    implements_by_view[view.view].add(implements)
                else:
                    warnings.warn(
                        NeatValueWarning(
                            f"Implemented view {implements} of view {view} is not in the same namespace, skipping!"
                        ),
                        stacklevel=2,
                    )
        if include_ancestors:
            self._include_ancestors(implements_by_view)
        return implements_by_view

    def properties_by_view(
        self, include_ancestors: bool = False, include_different_space: bool = False
    ) -> dict[ViewEntity, list[DMSProperty]]:
        """Get a dictionary of views and their properties."""
        # This is a duplicate fo the properties_by_class method, but for views
        # The choice to duplicate the code is to avoid generics which will make the code less readable.
        properties_by_views: dict[ViewEntity, list[DMSProperty]] = defaultdict(list)
        for prop in self.dms.properties:
            properties_by_views[prop.view].append(prop)

        if include_ancestors:
            implements_by_view = self.implements_by_view(
                include_ancestors=include_ancestors, include_different_space=include_different_space
            )
            for view, parents in implements_by_view.items():
                view_properties = {prop.view_property for prop in properties_by_views[view]}
                for parent in parents:
                    for parent_prop in properties_by_views[parent]:
                        if parent_prop.view_property not in view_properties:
                            child_prop = parent_prop.model_copy(update={"view": view})
                            properties_by_views[view].append(child_prop)
                            view_properties.add(child_prop.view_property)

        return properties_by_views

    @property
    def logical_uri_by_view(self) -> dict[ViewEntity, URIRef]:
        """Get the logical URI by view."""
        return {view.view: view.logical for view in self.dms.views if view.logical}

    def logical_uri_by_property_by_view(
        self,
        include_ancestors: bool = False,
        include_different_space: bool = False,
    ) -> dict[ViewEntity, dict[str, URIRef]]:
        """Get the logical URI by property by view."""
        properties_by_view = self.properties_by_view(include_ancestors, include_different_space)

        return {
            view: {prop.view_property: prop.logical for prop in properties if prop.logical}
            for view, properties in properties_by_view.items()
        }

    @property
    def _class_by_neat_id(self) -> dict[URIRef, InformationClass]:
        """Get a dictionary of class neat IDs to
        class entities."""

        return {cls.neatId: cls for cls in self.information.classes if cls.neatId}

    def class_by_suffix(self) -> dict[str, InformationClass]:
        """Get a dictionary of class suffixes to class entities."""
        # TODO: Remove this method
        class_dict: dict[str, InformationClass] = {}
        for definition in self.information.classes:
            entity = definition.class_
            if entity.suffix in class_dict:
                warnings.warn(
                    NeatValueWarning(
                        f"Class {entity} has been defined more than once! Only the first definition "
                        "will be considered, skipping the rest.."
                    ),
                    stacklevel=2,
                )
                continue
            class_dict[entity.suffix] = definition
        return class_dict

    @property
    def class_by_class_entity(self) -> dict[ClassEntity, InformationClass]:
        """Get a dictionary of class entities to class entities."""
        rules = self.information
        return {cls.class_: cls for cls in rules.classes}

    @property
    def view_by_view_entity(self) -> dict[ViewEntity, DMSView]:
        """Get a dictionary of class entities to class entities."""
        rules = self.dms
        return {view.view: view for view in rules.views}

    def property_by_id(self) -> dict[str, list[InformationProperty]]:
        """Get a dictionary of property IDs to property entities."""
        property_dict: dict[str, list[InformationProperty]] = defaultdict(list)
        for prop in self.information.properties:
            property_dict[prop.property_].append(prop)
        return property_dict

    def properties_by_id_by_class(
        self,
        has_instance_source: bool = False,
        include_ancestors: bool = False,
    ) -> dict[ClassEntity, dict[str, InformationProperty]]:
        """Get a dictionary of class entities to dictionaries of property IDs to property entities."""
        class_property_pairs: dict[ClassEntity, dict[str, InformationProperty]] = {}
        for class_, properties in self.properties_by_class(include_ancestors).items():
            processed_properties: dict[str, InformationProperty] = {}
            for prop in properties:
                if prop.property_ in processed_properties:
                    warnings.warn(
                        NeatValueWarning(
                            f"Property {processed_properties} for {class_} has been defined more than once!"
                            " Only the first definition will be considered, skipping the rest.."
                        ),
                        stacklevel=2,
                    )
                    continue
                if has_instance_source and prop.instance_source is None:
                    continue
                processed_properties[prop.property_] = prop
            class_property_pairs[class_] = processed_properties

        return class_property_pairs

    def defined_views(self, include_ancestors: bool = False) -> set[ViewEntity]:
        properties_by_view = self.properties_by_view(include_ancestors)
        return {prop.view for prop in itertools.chain.from_iterable(properties_by_view.values())}

    def defined_classes(
        self,
        include_ancestors: bool = False,
    ) -> set[ClassEntity]:
        """Returns classes that have properties defined for them in the data model.

        Args:
            include_ancestors: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of classes that have been defined in the data model
        """
        properties_by_class = self.properties_by_class(include_ancestors)
        return {prop.class_ for prop in itertools.chain.from_iterable(properties_by_class.values())}

    def class_linkage(
        self,
        include_ancestors: bool = False,
    ) -> LinkageSet:
        """Returns a set of class linkages in the data model.

        Args:
            include_ancestors: Whether to consider inheritance or not. Defaults False

        Returns:

        """
        class_linkage = LinkageSet()

        properties_by_class = self.properties_by_class(include_ancestors)

        prop: InformationProperty
        for prop in itertools.chain.from_iterable(properties_by_class.values()):
            if not isinstance(prop.value_type, ClassEntity):
                continue
            class_linkage.add(
                Linkage(
                    source_class=prop.class_,
                    connecting_property=prop.property_,
                    target_class=prop.value_type,
                    max_occurrence=prop.max_count,
                )
            )

        return class_linkage

    def symmetrically_connected_classes(
        self,
        include_ancestors: bool = False,
    ) -> set[tuple[ClassEntity, ClassEntity]]:
        """Returns a set of pairs of symmetrically linked classes.

        Args:
            include_ancestors: Whether to consider inheritance or not. Defaults False

        Returns:
            Set of pairs of symmetrically linked classes

        !!! note "Symmetrically Connected Classes"
            Symmetrically connected classes are classes that are connected to each other
            in both directions. For example, if class A is connected to class B, and class B
            is connected to class A, then classes A and B are symmetrically connected.
        """
        sym_pairs: set[tuple[ClassEntity, ClassEntity]] = set()
        class_linkage = self.class_linkage(include_ancestors)
        if not class_linkage:
            return sym_pairs

        targets_by_source = class_linkage.get_target_classes_by_source()
        for link in class_linkage:
            source = link.source_class
            target = link.target_class

            if source in targets_by_source[source] and (source, target) not in sym_pairs:
                sym_pairs.add((source, target))
        return sym_pairs

    @overload
    def _properties_by_neat_id(self, format: Literal["info"] = "info") -> dict[URIRef, InformationProperty]: ...

    @overload
    def _properties_by_neat_id(self, format: Literal["dms"] = "dms") -> dict[URIRef, DMSProperty]: ...

    def _properties_by_neat_id(
        self, format: Literal["info", "dms"] = "info"
    ) -> dict[URIRef, InformationProperty] | dict[URIRef, DMSProperty]:
        if format == "info":
            return {prop.neatId: prop for prop in self.information.properties if prop.neatId}
        elif format == "dms":
            return {prop.neatId: prop for prop in self.dms.properties if prop.neatId}
        else:
            raise NeatValueError(f"Invalid format: {format}")

    @property
    def classes_by_neat_id(self) -> dict[URIRef, InformationClass]:
        return {class_.neatId: class_ for class_ in self.information.classes if class_.neatId}

    @property
    def multi_value_properties(self) -> list[InformationProperty]:
        return [prop_ for prop_ in self.information.properties if isinstance(prop_.value_type, MultiValueTypeInfo)]

    @property
    def view_query_by_id(
        self,
    ) -> "ViewQueryDict":
        # Trigger error if any of these are missing
        _ = self.information
        _ = self.dms

        # caching results for faster access
        classes_by_neat_id = self._class_by_neat_id
        properties_by_class = self.properties_by_class(include_ancestors=True)
        logical_uri_by_view = self.logical_uri_by_view
        logical_uri_by_property_by_view = self.logical_uri_by_property_by_view(include_ancestors=True)
        information_properties_by_neat_id = self._properties_by_neat_id()

        query_configs = ViewQueryDict()
        for view in self.dms.views:
            # this entire block of sequential if statements checks:
            # 1. connection of dms to info rules
            # 2. correct paring of information and dms rules
            # 3. connection of info rules to instances
            if (
                (neat_id := logical_uri_by_view.get(view.view))
                and (class_ := classes_by_neat_id.get(neat_id))
                and (uri := class_.instance_source)
            ):
                view_query = ViewQuery(
                    view_id=view.view.as_id(),
                    rdf_type=uri,
                    # start off with renaming of properties on the information level
                    # this is to encounter for special cases of e.g. space, startNode and endNode
                    property_renaming_config=(
                        {uri: prop_.property_ for prop_ in info_properties for uri in prop_.instance_source or []}
                        if (info_properties := properties_by_class.get(class_.class_))
                        else {}
                    ),
                )

                if logical_uri_by_property := logical_uri_by_property_by_view.get(view.view):
                    for target_name, neat_id in logical_uri_by_property.items():
                        if (property_ := information_properties_by_neat_id.get(neat_id)) and (
                            uris := property_.instance_source
                        ):
                            for uri in uris:
                                view_query.property_renaming_config[uri] = target_name

                query_configs[view.view.as_id()] = view_query

        return query_configs

    def _dms_di_graph(self, format: Literal["data-model", "implements"] = "data-model") -> nx.DiGraph:
        """Generate a DiGraph from the DMS rules."""
        di_graph = nx.DiGraph()

        rules = self.dms

        # Add nodes and edges from Views sheet
        for view in rules.views:
            di_graph.add_node(view.view.suffix, label=view.view.suffix)

            if format == "implements" and view.implements:
                for implement in view.implements:
                    di_graph.add_node(implement.suffix, label=implement.suffix)
                    di_graph.add_edge(
                        view.view.suffix,
                        implement.suffix,
                        label="implements",
                        dashes=True,
                    )

        if format == "data-model":
            # Add nodes and edges from Properties sheet
            for prop_ in rules.properties:
                if prop_.connection and isinstance(prop_.value_type, ViewEntity):
                    di_graph.add_node(prop_.view.suffix, label=prop_.view.suffix)
                    di_graph.add_node(prop_.value_type.suffix, label=prop_.value_type.suffix)
                    di_graph.add_edge(
                        prop_.view.suffix,
                        prop_.value_type.suffix,
                        label=prop_.name or prop_.view_property,
                    )

        return di_graph

    def _info_di_graph(self, format: Literal["data-model", "implements"] = "data-model") -> nx.DiGraph:
        """Generate DiGraph representing information data model."""

        rules = self.information
        di_graph = nx.DiGraph()

        # Add nodes and edges from Views sheet
        for class_ in rules.classes:
            # if possible use human readable label coming from the view name

            di_graph.add_node(
                class_.class_.suffix,
                label=class_.name or class_.class_.suffix,
            )

            if format == "implements" and class_.implements:
                for parent in class_.implements:
                    di_graph.add_node(parent.suffix, label=parent.suffix)
                    di_graph.add_edge(
                        class_.class_.suffix,
                        parent.suffix,
                        label="implements",
                        dashes=True,
                    )

        if format == "data-model":
            # Add nodes and edges from Properties sheet
            for prop_ in rules.properties:
                if isinstance(prop_.value_type, ClassEntity) and not isinstance(prop_.value_type, UnknownEntity):
                    di_graph.add_node(prop_.class_.suffix, label=prop_.class_.suffix)
                    di_graph.add_node(prop_.value_type.suffix, label=prop_.value_type.suffix)

                    di_graph.add_edge(
                        prop_.class_.suffix,
                        prop_.value_type.suffix,
                        label=prop_.name or prop_.property_,
                    )

        return di_graph



import warnings
from collections.abc import Callable, Iterator
from typing import Any, cast
from urllib.parse import quote

import rdflib
from rdflib import RDF, RDFS, Literal, Namespace, URIRef
from rdflib.query import ResultRow

from cognite.neat._constants import NEAT
from cognite.neat._issues.warnings import PropertyDataTypeConversionWarning
from cognite.neat._utils.auxiliary import string_to_ideal_type
from cognite.neat._utils.rdf_ import Triple, get_namespace, remove_namespace_from_uri

from ._base import BaseTransformerStandardised, RowTransformationOutput


class SplitMultiValueProperty(BaseTransformerStandardised):
    description: str = (
        "SplitMultiValueProperty is a transformer that splits a "
        "multi-value property into multiple single-value properties."
    )
    _use_only_once: bool = True
    _need_changes = frozenset({})

    def __init__(self, unknown_type: URIRef | None = None) -> None:
        self.unknown_type = unknown_type or NEAT.UnknownType

    def _iterate_query(self) -> str:
        query = """SELECT ?subjectType ?property
                          (GROUP_CONCAT(DISTINCT STR(?valueType); SEPARATOR=",") AS ?valueTypes)

                   WHERE {{
                       ?s ?property ?o .
                       ?s a ?subjectType .
                       OPTIONAL {{ ?o a ?type }}

                       # Key part to determine value type: either object, data or unknown
                       BIND(   IF(isLiteral(?o),DATATYPE(?o),
                               IF(BOUND(?type), ?type,
                                               <{unknownType}>)) AS ?valueType)
                   }}

                   GROUP BY ?subjectType ?property
                   HAVING (COUNT(DISTINCT ?valueType) > 1)"""

        return query.format(unknownType=self.unknown_type)

    def _count_query(self) -> str:
        query = """SELECT (COUNT(*) AS ?tripleCount)
                   WHERE {?s ?p ?o .}"""
        return query

    def _sub_iterate_query(self, type_: URIRef, property_: URIRef) -> str:
        query = """ SELECT ?s ?p ?o ?valueType WHERE {{
                           ?s a <{subject_uri}> .
                           ?s <{property_uri}> ?o .

                           OPTIONAL {{ ?o a ?type }}

                           BIND(<{property_uri}> AS ?p)

                           BIND(IF(isLiteral(?o),  DATATYPE(?o),
                                   IF(BOUND(?type),?type,
                                                   <{unknownType}>)) AS ?valueType)

                                               }} """

        return query.format(unknownType=self.unknown_type, subject_uri=type_, property_uri=property_)

    def _iterator(self, graph) -> Iterator:
        for type_, property_, _ in graph.query(self._iterate_query()):
            yield from graph.query(self._sub_iterate_query(type_, property_))

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()
        subject, old_property, object, value_type = query_result_row

        new_property = URIRef(f"{old_property}_{remove_namespace_from_uri(value_type)}")

        row_output.add_triples.add(cast(Triple, (subject, new_property, object)))
        row_output.remove_triples.add(cast(Triple, (subject, old_property, object)))

        row_output.instances_modified_count += 1

        return row_output


class ConvertLiteral(BaseTransformerStandardised):
    description: str = "ConvertLiteral is a transformer that improve data typing of a literal value."
    _use_only_once: bool = False
    _need_changes = frozenset({})

    def __init__(
        self,
        subject_type: URIRef,
        subject_predicate: URIRef,
        conversion: Callable[[Any], Any] | None = None,
    ) -> None:
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate
        self.conversion = conversion or string_to_ideal_type
        self._type_name = remove_namespace_from_uri(subject_type)
        self._property_name = remove_namespace_from_uri(subject_predicate)

    def _skip_count_query(self) -> str:
        query = """SELECT (COUNT(?value) AS ?valueCount)
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?value
                       FILTER(isIRI(?value))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _count_query(self) -> str:
        query = """SELECT (COUNT(?value) AS ?valueCount)
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?value
                       FILTER(isLiteral(?value))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _iterate_query(self) -> str:
        query = """SELECT ?instance ?value
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?value
                      FILTER(isLiteral(?value))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        instance, literal = query_result_row
        value = cast(rdflib.Literal, literal).toPython()

        try:
            converted_value = self.conversion(value)
        except Exception as e:
            warnings.warn(
                PropertyDataTypeConversionWarning(str(instance), self._type_name, self._property_name, str(e)),
                stacklevel=2,
            )
        row_output.add_triples.add((instance, self.subject_predicate, rdflib.Literal(converted_value)))  # type: ignore[arg-type]
        row_output.remove_triples.add((instance, self.subject_predicate, literal))  # type: ignore[arg-type]
        row_output.instances_modified_count += 1

        return row_output


class LiteralToEntity(BaseTransformerStandardised):
    description = "Converts a literal value to new entity"

    def __init__(
        self, subject_type: URIRef | None, subject_predicate: URIRef, entity_type: str, new_property: str | None = None
    ) -> None:
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate
        self.entity_type = entity_type
        self.new_property = new_property

    def _iterate_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT ?instance ?property
            WHERE {{
              ?instance <{subject_predicate}> ?property
              FILTER(isLiteral(?property))
            }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT ?instance ?property
                WHERE {{
                  ?instance a <{subject_type}> .
                  ?instance <{subject_predicate}> ?property
                  FILTER(isLiteral(?property))
                }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _skip_count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                        WHERE {{
                          ?instance <{subject_predicate}> ?property
                          FILTER(isIRI(?property))
                        }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?property
                          FILTER(isIRI(?property))
                        }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                WHERE {{
                  ?instance <{subject_predicate}> ?property
                  FILTER(isLiteral(?property))
                }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?property) AS ?propertyCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?property
                          FILTER(isLiteral(?property))
                        }}"""

            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        instance, literal = query_result_row
        value = cast(rdflib.Literal, literal).toPython()
        namespace = Namespace(get_namespace(instance))  # type: ignore[arg-type]
        entity_type = namespace[self.entity_type]
        new_entity = namespace[f"{self.entity_type}_{quote(value)!s}"]
        row_output.add_triples.add((new_entity, RDF.type, entity_type))
        row_output.instances_added_count += 1  # we add one new entity

        if self.new_property is not None:
            row_output.add_triples.add((new_entity, namespace[self.new_property], rdflib.Literal(value)))  # type: ignore[arg-type]
            row_output.instances_modified_count += 1  # we modify the new entity

        row_output.add_triples.add((instance, self.subject_predicate, new_entity))  # type: ignore[arg-type]
        row_output.remove_triples.add((instance, self.subject_predicate, literal))  # type: ignore[arg-type]
        row_output.instances_modified_count += 1  # we modify the old entity

        return row_output


class ConnectionToLiteral(BaseTransformerStandardised):
    description = "Converts an entity connection to a literal value"

    def __init__(self, subject_type: URIRef | None, subject_predicate: URIRef) -> None:
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate

    def _iterate_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT ?instance ?object
            WHERE {{
              ?instance <{subject_predicate}> ?object
              FILTER(isIRI(?object))
            }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT ?instance ?object
                WHERE {{
                  ?instance a <{subject_type}> .
                  ?instance <{subject_predicate}> ?object
                  FILTER(isIRI(?object))
                }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _skip_count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                        WHERE {{
                          ?instance <{subject_predicate}> ?object
                          FILTER(isLiteral(?object))
                        }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?object
                          FILTER(isLiteral(?object))
                        }}"""
            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _count_query(self) -> str:
        if self.subject_type is None:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                WHERE {{
                  ?instance <{subject_predicate}> ?object
                  FILTER(isIRI(?object))
                }}"""
            return query.format(subject_predicate=self.subject_predicate)
        else:
            query = """SELECT (COUNT(?object) AS ?objectCount)
                        WHERE {{
                          ?instance a <{subject_type}> .
                          ?instance <{subject_predicate}> ?object
                          FILTER(isIRI(?object))
                        }}"""

            return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        instance, object_entity = cast(tuple[URIRef, URIRef], query_result_row)
        value = remove_namespace_from_uri(object_entity)

        row_output.add_triples.add((instance, self.subject_predicate, rdflib.Literal(value)))
        row_output.remove_triples.add((instance, self.subject_predicate, object_entity))
        row_output.instances_modified_count += 1

        return row_output


class SetType(BaseTransformerStandardised):
    description = "Set the type of an instance based on a property"

    def __init__(
        self,
        subject_type: URIRef,
        subject_predicate: URIRef,
        drop_property: bool = False,
        namespace: Namespace | None = None,
    ) -> None:
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate
        self.drop_property = drop_property
        self._namespace = namespace or Namespace(get_namespace(subject_type))

    def _count_query(self) -> str:
        query = """SELECT (COUNT(?object) AS ?objectCount)
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?object
                      FILTER(isLiteral(?object))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _skip_count_query(self) -> str:
        query = """SELECT (COUNT(?object) AS ?objectCount)
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?object
                      FILTER(isIRI(?object))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def _iterate_query(self) -> str:
        query = """SELECT ?instance ?object
                    WHERE {{
                      ?instance a <{subject_type}> .
                      ?instance <{subject_predicate}> ?object
                      FILTER(isLiteral(?object))
                    }}"""
        return query.format(subject_type=self.subject_type, subject_predicate=self.subject_predicate)

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        instance, object_literal = cast(tuple[URIRef, Literal], query_result_row)
        if self.drop_property:
            row_output.remove_triples.add((instance, self.subject_predicate, object_literal))

        row_output.remove_triples.add((instance, RDF.type, self.subject_type))
        new_type = self._namespace[quote(object_literal.toPython())]
        row_output.add_triples.add((instance, RDF.type, new_type))
        row_output.add_triples.add((new_type, RDFS.subClassOf, self.subject_type))
        row_output.instances_modified_count += 1

        return row_output



import itertools
from collections import Counter, defaultdict
from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, ClassVar, cast

from cognite.client import data_modeling as dm
from rdflib import RDF, RDFS, Graph, Namespace, URIRef
from rdflib import Literal as RdfLiteral

from cognite.neat._config import GLOBAL_CONFIG
from cognite.neat._constants import NEAT, get_default_prefixes_and_namespaces
from cognite.neat._issues import IssueList
from cognite.neat._issues.warnings import PropertyValueTypeUndefinedWarning
from cognite.neat._rules.analysis import RulesAnalysis
from cognite.neat._rules.models import InformationRules, data_types
from cognite.neat._rules.models.data_types import AnyURI
from cognite.neat._rules.models.entities._single_value import UnknownEntity
from cognite.neat._rules.models.information import (
    InformationClass,
    InformationInputClass,
    InformationInputProperty,
    InformationMetadata,
)
from cognite.neat._store import NeatGraphStore
from cognite.neat._store._provenance import INSTANCES_ENTITY
from cognite.neat._utils.collection_ import iterate_progress_bar
from cognite.neat._utils.rdf_ import remove_namespace_from_uri, uri_to_short_form
from cognite.neat._utils.text import NamingStandardization

from ._base import DEFAULT_NON_EXISTING_NODE_TYPE, BaseRDFImporter

DEFAULT_INFERENCE_DATA_MODEL_ID = ("neat_space", "InferredDataModel", "inferred")

ORDERED_CLASSES_QUERY = """SELECT ?class (count(?s) as ?instances )
                           WHERE { ?s a ?class . }
                           group by ?class order by DESC(?instances)"""

INSTANCES_OF_CLASS_QUERY = """SELECT ?s ?propertyCount WHERE { ?s a <class> . BIND ('Unknown' as ?propertyCount) }"""


INSTANCES_OF_CLASS_RICHNESS_ORDERED_QUERY = """SELECT ?s (COUNT(?p) as ?propertyCount)
                                               WHERE { ?s a <class> ; ?p ?o . }
                                               GROUP BY ?s
                                               ORDER BY DESC(?propertyCount)"""

INSTANCE_PROPERTIES_DEFINITION = """SELECT ?property (count(?property) as ?occurrence) ?dataType ?objectType
                                    WHERE {<instance_id> ?property ?value .

                                           BIND(datatype(?value) AS ?dataType)

                                           OPTIONAL {?value rdf:type ?objectType .}}
                                    GROUP BY ?property ?dataType ?objectType"""


class InferenceImporter(BaseRDFImporter):
    """Infers rules from a triple store.

    Rules inference through analysis of knowledge graph provided in various formats.
    Use the factory methods to create a triple store from sources such as
    RDF files, JSON files, YAML files, XML files, or directly from a graph store.

    ClassVars:
        overwrite_data_types: Mapping of data types to be overwritten. The InferenceImporter will overwrite
            32-bit integer and 32-bit float data types to 64-bit integer and 64-bit float data types

    Args:
        issue_list: Issue list to store issues
        graph: Knowledge graph
        max_number_of_instance: Maximum number of instances to be used in inference
        prefix: Prefix to be used for the inferred model


    """

    overwrite_data_types: ClassVar[Mapping[URIRef, URIRef]] = {
        data_types.Integer.as_xml_uri_ref(): data_types.Long.as_xml_uri_ref(),
        data_types.Float.as_xml_uri_ref(): data_types.Double.as_xml_uri_ref(),
    }

    @classmethod
    def from_graph_store(
        cls,
        store: NeatGraphStore,
        data_model_id: dm.DataModelId | tuple[str, str, str] = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
        language: str = "en",
    ) -> "InferenceImporter":
        return super().from_graph_store(
            store,
            data_model_id,
            max_number_of_instance,
            non_existing_node_type,
            language,
        )

    @classmethod
    def from_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
        language: str = "en",
        source_name: str = "Unknown",
    ) -> "InferenceImporter":
        return super().from_file(
            filepath,
            data_model_id,
            max_number_of_instance,
            non_existing_node_type,
            language,
            source_name=source_name,
        )

    @classmethod
    def from_json_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        language: str = "en",
    ) -> "InferenceImporter":
        raise NotImplementedError("JSON file format is not supported yet.")

    @classmethod
    def from_yaml_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
        language: str = "en",
    ) -> "InferenceImporter":
        raise NotImplementedError("YAML file format is not supported yet.")

    @classmethod
    def from_xml_file(
        cls,
        filepath: Path,
        data_model_id: (dm.DataModelId | tuple[str, str, str]) = DEFAULT_INFERENCE_DATA_MODEL_ID,
        max_number_of_instance: int = -1,
    ) -> "InferenceImporter":
        raise NotImplementedError("JSON file format is not supported yet.")

    def _to_rules_components(
        self,
    ) -> dict:
        """Convert RDF graph to dictionary defining data model and prefixes of the graph

        Args:
            graph: RDF graph to be converted to TransformationRules object
            max_number_of_instance: Max number of instances to be considered for each class

        Returns:
            Tuple of data model and prefixes of the graph
        """

        classes: dict[str, dict] = {}
        properties: dict[str, dict] = {}
        prefixes: dict[str, Namespace] = {}
        count_by_value_type_by_property: dict[str, dict[str, int]] = defaultdict(Counter)

        # Infers all the classes in the graph
        for class_uri, no_instances in self.graph.query(ORDERED_CLASSES_QUERY):  # type: ignore[misc]
            if (class_id := remove_namespace_from_uri(cast(URIRef, class_uri))) in classes:
                # handles cases when class id is already present in classes
                class_id = f"{class_id}_{len(classes) + 1}"

            classes[class_id] = {
                "class_": class_id,
                "uri": class_uri,
                "comment": f"Inferred from knowledge graph, where this class has <{no_instances}> instances",
            }

            self._add_uri_namespace_to_prefixes(cast(URIRef, class_uri), prefixes)

        instances_query = (
            INSTANCES_OF_CLASS_QUERY if self.max_number_of_instance == -1 else INSTANCES_OF_CLASS_RICHNESS_ORDERED_QUERY
        )

        classes_iterable = iterate_progress_bar(classes.items(), len(classes), "Inferring classes")

        # Infers all the properties of the class
        for class_id, class_definition in classes_iterable:
            for (  # type: ignore[misc]
                instance,
                _,
            ) in self.graph.query(  # type: ignore[misc]
                instances_query.replace("class", class_definition["uri"])
                if self.max_number_of_instance < 0
                else instances_query.replace("class", class_definition["uri"]) + f" LIMIT {self.max_number_of_instance}"
            ):
                for property_uri, occurrence, data_type_uri, object_type_uri in self.graph.query(  # type: ignore[misc]
                    INSTANCE_PROPERTIES_DEFINITION.replace("instance_id", instance)
                ):  # type: ignore[misc]
                    # this is to skip rdf:type property

                    if property_uri == RDF.type:
                        continue
                    property_id = remove_namespace_from_uri(property_uri)
                    self._add_uri_namespace_to_prefixes(cast(URIRef, property_uri), prefixes)

                    if isinstance(data_type_uri, URIRef):
                        data_type_uri = self.overwrite_data_types.get(data_type_uri, data_type_uri)

                    if value_type_uri := (data_type_uri or object_type_uri):
                        self._add_uri_namespace_to_prefixes(cast(URIRef, value_type_uri), prefixes)
                        value_type_id = remove_namespace_from_uri(value_type_uri)

                    # this handles situations when property points to node that is not present in graph
                    else:
                        value_type_id = str(self.non_existing_node_type)

                        issue = PropertyValueTypeUndefinedWarning(
                            resource_type="Property",
                            identifier=f"{class_id}:{property_id}",
                            property_name=property_id,
                            default_action="Remove the property from the rules",
                            recommended_action="Make sure that graph is complete",
                        )

                        if issue not in self.issue_list:
                            self.issue_list.append(issue)

                    id_ = f"{class_id}:{property_id}"

                    definition = {
                        "class_": class_id,
                        "property_": property_id,
                        "max_count": cast(RdfLiteral, occurrence).value,
                        "value_type": value_type_id,
                        "instance_source": (
                            f"{uri_to_short_form(class_definition['uri'], prefixes)}"
                            f"({uri_to_short_form(cast(URIRef, property_uri), prefixes)})"
                        ),
                    }

                    count_by_value_type_by_property[id_][value_type_id] += 1

                    # USE CASE 1: If property is not present in properties
                    if id_ not in properties:
                        definition["value_type"] = {definition["value_type"]}
                        properties[id_] = definition

                    # USE CASE 2: first time redefinition, value type change to multi
                    elif id_ in properties and definition["value_type"] not in properties[id_]["value_type"]:
                        properties[id_]["value_type"].add(definition["value_type"])

                    # always update max_count with the upmost value
                    properties[id_]["max_count"] = max(properties[id_]["max_count"], definition["max_count"])

        # Create multi-value properties otherwise single value
        for property_ in properties.values():
            # Removes non-existing node type from value type prior final conversion to string
            if len(property_["value_type"]) > 1 and str(self.non_existing_node_type) in property_["value_type"]:
                property_["value_type"].remove(str(self.non_existing_node_type))

            if len(property_["value_type"]) > 1:
                property_["value_type"] = ", ".join([str(t) for t in property_["value_type"]])
            else:
                property_["value_type"] = next(iter(property_["value_type"]))

        return {
            "metadata": self._default_metadata().model_dump(),
            "classes": list(classes.values()),
            "properties": list(properties.values()),
            "prefixes": prefixes,
        }

    def _default_metadata(self):
        now = datetime.now(timezone.utc)
        return InformationMetadata(
            space=self.data_model_id.space,
            external_id=self.data_model_id.external_id,
            version=self.data_model_id.version,
            name="Inferred Model",
            creator="NEAT",
            created=now,
            updated=now,
            description="Inferred model from knowledge graph",
        )

    @property
    def source_uri(self) -> URIRef:
        return INSTANCES_ENTITY.id_


# Internal helper class
@dataclass
class _ReadProperties:
    type_uri: URIRef
    property_uri: URIRef
    value_type: URIRef
    parent_uri: URIRef | None
    max_occurrence: int
    instance_count: int


class SubclassInferenceImporter(BaseRDFImporter):
    """Infer subclasses from a triple store.

    Assumes that the graph already is connected to a schema. The classes should
    match the RDF.type of the instances in the graph, while the subclasses should
    match the NEAT.type of the instances in the graph.

    ClassVars:
        overwrite_data_types: Mapping of data types to be overwritten. The InferenceImporter will overwrite
            32-bit integer and 32-bit float data types to 64-bit integer and 64-bit float data types

    Args:
        issue_list: Issue list to store issues
        graph: Knowledge graph
    """

    overwrite_data_types: ClassVar[Mapping[URIRef, URIRef]] = {
        data_types.Integer.as_xml_uri_ref(): data_types.Long.as_xml_uri_ref(),
        data_types.Float.as_xml_uri_ref(): data_types.Double.as_xml_uri_ref(),
    }

    _ordered_class_query = """SELECT DISTINCT ?class (count(?s) as ?instances )
                           WHERE { ?s a ?class }
                           group by ?class order by DESC(?instances)"""

    _type_parent_query = f"""SELECT ?parent ?type
                            WHERE {{ ?s a ?type .
                            ?type <{RDFS.subClassOf}> ?parent }}"""

    _properties_query = """SELECT DISTINCT ?property ?valueType
                         WHERE {{
                            ?s a <{type}> .
                            ?s ?property ?object .
                            OPTIONAL {{ ?object a ?objectType }}
                            BIND(
                               IF(
                                    isLiteral(?object), datatype(?object),
                                    IF(BOUND(?objectType), ?objectType, <{unknown_type}>)
                                ) AS ?valueType
                            )
                        }}"""

    _max_occurrence_query = """SELECT (MAX(?count) AS ?maxCount)
                            WHERE {{
                              {{
                                SELECT ?subject (COUNT(?object) AS ?count)
                                WHERE {{
                                  ?subject a <{type}> .
                                  ?subject <{property}> ?object .
                                }}
                                GROUP BY ?subject
                              }}
                            }}"""

    def __init__(
        self,
        issue_list: IssueList,
        graph: Graph,
        rules: InformationRules | None = None,
        data_model_id: dm.DataModelId | tuple[str, str, str] | None = None,
        non_existing_node_type: UnknownEntity | AnyURI = DEFAULT_NON_EXISTING_NODE_TYPE,
    ) -> None:
        if sum([1 for v in [rules, data_model_id] if v is not None]) != 1:
            raise ValueError("Exactly one of rules or data_model_id must be provided.")
        if data_model_id is not None:
            identifier = data_model_id
        elif rules is not None:
            identifier = rules.metadata.as_data_model_id().as_tuple()  # type: ignore[assignment]
        else:
            raise ValueError("Exactly one of rules or data_model_id must be provided.")
        super().__init__(issue_list, graph, identifier, -1, non_existing_node_type, language="en")
        self._rules = rules

    def _to_rules_components(
        self,
    ) -> dict:
        if self._rules:
            prefixes = self._rules.prefixes.copy()
        else:
            prefixes = get_default_prefixes_and_namespaces()

        parent_by_child = self._read_parent_by_child_from_graph()
        read_properties = self._read_class_properties_from_graph(parent_by_child)
        classes, properties = self._create_classes_properties(read_properties, prefixes)

        if self._rules:
            metadata = self._rules.metadata.model_dump()
            default_space = self._rules.metadata.prefix
        else:
            metadata = self._default_metadata()
            default_space = metadata["space"]
        return {
            "metadata": metadata,
            "classes": [cls.dump(default_space) for cls in classes],
            "properties": [prop.dump(default_space) for prop in properties],
            "prefixes": prefixes,
        }

    def _create_classes_properties(
        self, read_properties: list[_ReadProperties], prefixes: dict[str, Namespace]
    ) -> tuple[list[InformationInputClass], list[InformationInputProperty]]:
        if self._rules:
            existing_classes = {class_.class_.suffix: class_ for class_ in self._rules.classes}
        else:
            existing_classes = {}
        classes: list[InformationInputClass] = []
        properties_by_class_suffix_by_property_id: dict[str, dict[str, InformationInputProperty]] = {}

        # Help for IDE
        type_uri: URIRef
        parent_uri: URIRef
        for parent_uri, parent_class_properties_iterable in itertools.groupby(
            sorted(read_properties, key=lambda x: x.parent_uri or NEAT.EmptyType),
            key=lambda x: x.parent_uri or NEAT.EmptyType,
        ):
            properties_by_class_by_property = self._get_properties_by_class_by_property(
                parent_class_properties_iterable
            )

            parent_suffix: str | None = None
            if parent_uri != NEAT.EmptyType:
                shared_property_uris = set.intersection(
                    *[
                        set(properties_by_property.keys())
                        for properties_by_property in properties_by_class_by_property.values()
                    ]
                )
                parent_suffix = remove_namespace_from_uri(parent_uri)
                self._add_uri_namespace_to_prefixes(parent_uri, prefixes)
                if parent_suffix not in existing_classes:
                    classes.append(InformationInputClass(class_=parent_suffix))
                else:
                    classes.append(InformationInputClass.load(existing_classes[parent_suffix].model_dump()))
            else:
                shared_property_uris = set()
            shared_properties: dict[URIRef, list[_ReadProperties]] = defaultdict(list)
            for type_uri, properties_by_property_uri in properties_by_class_by_property.items():
                class_suffix = remove_namespace_from_uri(type_uri)
                self._add_uri_namespace_to_prefixes(type_uri, prefixes)

                if class_suffix not in existing_classes:
                    classes.append(
                        InformationInputClass(
                            class_=class_suffix,
                            implements=parent_suffix,
                            instance_source=type_uri,
                        )
                    )
                else:
                    classes.append(InformationInputClass.load(existing_classes[class_suffix].model_dump()))

                properties_by_id: dict[str, InformationInputProperty] = {}
                for property_uri, read_properties in properties_by_property_uri.items():
                    if property_uri in shared_property_uris:
                        shared_properties[property_uri].extend(read_properties)
                        continue
                    property_id = remove_namespace_from_uri(property_uri)
                    self._add_uri_namespace_to_prefixes(property_uri, prefixes)
                    property_id_standardized = NamingStandardization.standardize_property_str(property_uri)
                    if existing_prop := properties_by_id.get(property_id_standardized):
                        if not isinstance(existing_prop.instance_source, list):
                            existing_prop.instance_source = (
                                [existing_prop.instance_source] if existing_prop.instance_source else []
                            )
                        existing_prop.instance_source.append(property_uri)
                        continue
                    else:
                        properties_by_id[property_id_standardized] = self._create_property(
                            read_properties, class_suffix, property_uri, property_id, prefixes
                        )
                properties_by_class_suffix_by_property_id[class_suffix] = properties_by_id
            if parent_suffix:
                properties_by_id = {}
                for property_uri, read_properties in shared_properties.items():
                    property_id = remove_namespace_from_uri(property_uri)
                    self._add_uri_namespace_to_prefixes(property_uri, prefixes)
                    property_id_standardized = NamingStandardization.standardize_property_str(property_uri)
                    if existing_prop := properties_by_id.get(property_id_standardized):
                        if not isinstance(existing_prop.instance_source, list):
                            existing_prop.instance_source = (
                                [existing_prop.instance_source] if existing_prop.instance_source else []
                            )
                        existing_prop.instance_source.append(property_uri)
                    else:
                        properties_by_id[property_id_standardized] = self._create_property(
                            read_properties, parent_suffix, property_uri, property_id, prefixes
                        )
        return classes, [
            prop for properties in properties_by_class_suffix_by_property_id.values() for prop in properties.values()
        ]

    @staticmethod
    def _get_properties_by_class_by_property(
        parent_class_properties_iterable: Iterable[_ReadProperties],
    ) -> dict[URIRef, dict[URIRef, list[_ReadProperties]]]:
        properties_by_class_by_property: dict[URIRef, dict[URIRef, list[_ReadProperties]]] = {}
        for class_uri, class_properties_iterable in itertools.groupby(
            sorted(parent_class_properties_iterable, key=lambda x: x.type_uri), key=lambda x: x.type_uri
        ):
            properties_by_class_by_property[class_uri] = defaultdict(list)
            for read_prop in class_properties_iterable:
                properties_by_class_by_property[class_uri][read_prop.property_uri].append(read_prop)
        return properties_by_class_by_property

    def _read_class_properties_from_graph(self, parent_by_child: dict[URIRef, URIRef]) -> list[_ReadProperties]:
        count_by_type: dict[URIRef, int] = {}
        # Infers all the classes in the graph
        for result_row in self.graph.query(self._ordered_class_query):
            type_uri, instance_count_literal = cast(tuple[URIRef, RdfLiteral], result_row)
            count_by_type[type_uri] = instance_count_literal.toPython()
        if self._rules:
            analysis = RulesAnalysis(self._rules)
            existing_class_properties = {
                (class_entity.suffix, prop.property_): prop
                for class_entity, properties in analysis.properties_by_class(
                    include_ancestors=True, include_different_space=True
                ).items()
                for prop in properties
            }
            existing_classes = {cls_.class_.suffix: cls_ for cls_ in self._rules.classes}
        else:
            existing_class_properties = {}
            existing_classes = {}
        properties_by_class_by_subclass: list[_ReadProperties] = []
        existing_class: InformationClass | None
        total_instance_count = sum(count_by_type.values())
        iterable = count_by_type.items()
        if GLOBAL_CONFIG.use_iterate_bar_threshold and total_instance_count > GLOBAL_CONFIG.use_iterate_bar_threshold:
            iterable = iterate_progress_bar(iterable, len(count_by_type), "Inferring types...")  # type: ignore[assignment]
        for type_uri, instance_count in iterable:
            property_query = self._properties_query.format(type=type_uri, unknown_type=NEAT.UnknownType)
            class_suffix = remove_namespace_from_uri(type_uri)
            if (existing_class := existing_classes.get(class_suffix)) and existing_class.instance_source is None:
                existing_class.instance_source = type_uri

            for result_row in self.graph.query(property_query):
                property_uri, value_type_uri = cast(tuple[URIRef, URIRef], result_row)
                if property_uri == RDF.type:
                    continue
                property_str = remove_namespace_from_uri(property_uri)
                if existing_property := existing_class_properties.get((class_suffix, property_str)):
                    if existing_property.instance_source is None:
                        existing_property.instance_source = [property_uri]
                    elif existing_property.instance_source and property_uri not in existing_property.instance_source:
                        existing_property.instance_source.append(property_uri)
                    continue
                occurrence_query = self._max_occurrence_query.format(type=type_uri, property=property_uri)
                max_occurrence = 1  # default value
                occurrence_row, *_ = list(self.graph.query(occurrence_query))
                if occurrence_row:
                    max_occurrence_literal, *__ = cast(tuple[RdfLiteral, Any], occurrence_row)
                    max_occurrence = int(max_occurrence_literal.toPython())
                properties_by_class_by_subclass.append(
                    _ReadProperties(
                        type_uri=type_uri,
                        property_uri=property_uri,
                        parent_uri=parent_by_child.get(type_uri),
                        value_type=value_type_uri,
                        max_occurrence=max_occurrence,
                        instance_count=instance_count,
                    )
                )
        return properties_by_class_by_subclass

    def _read_parent_by_child_from_graph(self) -> dict[URIRef, URIRef]:
        parent_by_child: dict[URIRef, URIRef] = {}
        for result_row in self.graph.query(self._type_parent_query):
            parent_uri, child_uri = cast(tuple[URIRef, URIRef], result_row)
            parent_by_child[child_uri] = parent_uri
        return parent_by_child

    def _create_property(
        self,
        read_properties: list[_ReadProperties],
        class_suffix: str,
        property_uri: URIRef,
        property_id: str,
        prefixes: dict[str, Namespace],
    ) -> InformationInputProperty:
        first = read_properties[0]
        value_type = self._get_value_type(read_properties, prefixes)
        return InformationInputProperty(
            class_=class_suffix,
            property_=property_id,
            max_count=first.max_occurrence,
            value_type=value_type,
            instance_source=[property_uri],
        )

    def _get_value_type(
        self, read_properties: list[_ReadProperties], prefixes: dict[str, Namespace]
    ) -> str | UnknownEntity:
        value_types = {self.overwrite_data_types.get(prop.value_type, prop.value_type) for prop in read_properties}
        if len(value_types) == 1:
            uri_ref = value_types.pop()
            if uri_ref == NEAT.UnknownType:
                return UnknownEntity()
            self._add_uri_namespace_to_prefixes(uri_ref, prefixes)
            return remove_namespace_from_uri(uri_ref)
        elif len(value_types) == 0:
            return UnknownEntity()
        for uri_ref in value_types:
            self._add_uri_namespace_to_prefixes(uri_ref, prefixes)
        return ", ".join(remove_namespace_from_uri(uri_ref) for uri_ref in value_types)

    def _default_metadata(self) -> dict[str, Any]:
        now = datetime.now(timezone.utc)
        return InformationMetadata(
            space=self.data_model_id.space,
            external_id=self.data_model_id.external_id,
            version=cast(str, self.data_model_id.version),
            name="Inferred Model",
            creator=["NEAT"],
            created=now,
            updated=now,
            description="Inferred model from knowledge graph",
        ).model_dump()



from typing import Literal, TypeAlias

MIMETypes: TypeAlias = Literal[
    "application/rdf+xml", "text/turtle", "application/n-triple", "application/n-quads", "application/trig"
]

RDFTypes: TypeAlias = Literal["xml", "rdf", "owl", "n3", "ttl", "turtle", "nt", "nq", "nquads", "trig"]


def quad_formats() -> list[str]:
    return ["trig", "nquads", "nq", "nt"]


def rdflib_to_oxi_type(rdflib_format: str) -> str | None:
    """Convert an RDFlib format to a MIME type.

    Args:
        rdflib_format (str): The RDFlib format.

    Returns:
        Oxi format used to trigger correct plugging in rdflib

    """

    mapping = {
        "xml": "ox-xml",
        "rdf": "ox-xml",
        "owl": "ox-xml",
        "n3": "ox-n3",
        "ttl": "ox-ttl",
        "turtle": "ox-turtle",
        "nt": "ox-nt",
        "nq": "ox-nq",
        "nquads": "ox-nquads",
        "trig": "ox-trig",
    }
    return mapping.get(rdflib_format, None)



"""
We use prov-o to represent the provenance of instances and data models
basically tracking changes that occur.
prov-o use concepts of Agent, Activity and Entity to represent provenance
where in case of neat when dealing with instances we have:

 * Agent: triples extractors, graph enhancers, contextualizers, etc.
 * Activity: write/remove triples such as add connection, etc.
 * Entity: neat graph store

 and in case of data models we have:

 * Agent: Rules importers, exporters, transformers, etc.
 * Activity: convert, verify, etc.
 * Entity: data model (aka Rules)

"""

import uuid
from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field
from datetime import datetime
from typing import Generic, TypeVar

from rdflib import PROV, RDF, Literal, URIRef

from cognite.neat._constants import CDF_NAMESPACE, DEFAULT_NAMESPACE
from cognite.neat._issues import IssueList
from cognite.neat._shared import FrozenNeatObject, NeatList, Triple


@dataclass(frozen=True)
class Agent:
    id_: URIRef = DEFAULT_NAMESPACE.agent
    acted_on_behalf_of: str = "NEAT"

    def as_triples(self) -> list[Triple]:
        return [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.actedOnBehalfOf, Literal(self.acted_on_behalf_of)),
        ]


CDF_AGENT = Agent(acted_on_behalf_of="UNKNOWN", id_=CDF_NAMESPACE["agent"])
NEAT_AGENT = Agent(acted_on_behalf_of="UNKNOWN", id_=DEFAULT_NAMESPACE["agent"])
UNKNOWN_AGENT = Agent(acted_on_behalf_of="UNKNOWN", id_=DEFAULT_NAMESPACE["unknown-agent"])
EXTERNAL_AGENT = Agent(acted_on_behalf_of="HUMAN", id_=DEFAULT_NAMESPACE["external-agent"])


@dataclass(frozen=True)
class Entity:
    was_attributed_to: Agent
    issues: IssueList
    was_generated_by: "Activity | None" = field(repr=False)
    id_: URIRef

    @classmethod
    def create_with_defaults(
        cls,
        was_attributed_to: Agent,
        issues: IssueList | None = None,
        was_generated_by: "Activity | None" = None,
        id_: URIRef = DEFAULT_NAMESPACE["graph-store"],
    ) -> "Entity":
        return cls(
            was_attributed_to=was_attributed_to,
            issues=issues or IssueList(),
            was_generated_by=was_generated_by,
            id_=id_,
        )

    @classmethod
    def create_new_unknown_entity(cls) -> "Entity":
        return cls.create_with_defaults(
            was_attributed_to=UNKNOWN_AGENT,
            id_=DEFAULT_NAMESPACE[f"unknown-entity/{uuid.uuid4()}"],
        )

    def as_triples(self) -> list[Triple]:
        output: list[tuple[URIRef, URIRef, Literal | URIRef]] = [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.wasAttributedTo, self.was_attributed_to.id_),
        ]

        if self.was_generated_by:
            output.append(
                (
                    self.id_,
                    PROV.wasGeneratedBy,
                    self.was_generated_by.id_,
                )
            )

        return output


T_Entity = TypeVar("T_Entity", bound=Entity)
INSTANCES_ENTITY = Entity.create_with_defaults(was_attributed_to=NEAT_AGENT, id_=CDF_NAMESPACE["instances"])
EMPTY_ENTITY = Entity.create_with_defaults(was_attributed_to=NEAT_AGENT, id_=DEFAULT_NAMESPACE["empty-entity"])


@dataclass(frozen=True)
class Activity:
    was_associated_with: Agent
    ended_at_time: datetime
    started_at_time: datetime
    used: str | Entity | None = None
    id_: URIRef = field(default_factory=lambda: DEFAULT_NAMESPACE[f"activity-{uuid.uuid4()}"])

    def as_triples(self) -> list[Triple]:
        output: list[tuple[URIRef, URIRef, Literal | URIRef]] = [
            (self.id_, RDF.type, PROV[type(self).__name__]),
            (self.id_, PROV.wasAssociatedWith, self.was_associated_with.id_),
            (self.id_, PROV.startedAtTime, Literal(self.started_at_time)),
            (self.id_, PROV.endedAtTime, Literal(self.ended_at_time)),
        ]

        if self.used:
            output.append(
                (
                    self.id_,
                    PROV.used,
                    (self.used.id_ if isinstance(self.used, Entity) else Literal(self.used)),
                )
            )

        return output


@dataclass(frozen=True)
class Change(FrozenNeatObject, Generic[T_Entity]):
    agent: Agent
    activity: Activity
    target_entity: T_Entity
    description: str
    source_entity: Entity = field(default_factory=Entity.create_new_unknown_entity)

    def as_triples(self) -> list[Triple]:
        return (
            self.source_entity.as_triples()
            + self.agent.as_triples()
            + self.activity.as_triples()
            + self.target_entity.as_triples()  # type: ignore[operator]
        )

    @classmethod
    def record(cls, activity: str, start: datetime, end: datetime, description: str) -> "Change[Entity]":
        """User friendly method to record a change that occurred in the graph store."""
        agent = Agent()
        activity = Activity(
            used=activity,
            was_associated_with=agent,
            started_at_time=start,
            ended_at_time=end,
        )
        target_entity = Entity.create_with_defaults(was_generated_by=activity, was_attributed_to=agent)
        return Change(
            agent=agent,
            activity=activity,
            target_entity=target_entity,
            description=description,
        )

    def dump(self, aggregate: bool = True) -> dict[str, str]:
        return {
            "Source Entity": self.source_entity.id_,
            "Agent": self.agent.id_,
            "Activity": self.activity.id_,
            "Target Entity": self.target_entity.id_,
            "Description": self.description,
        }


class Provenance(NeatList[Change[T_Entity]]):
    def __init__(self, changes: Sequence[Change] | None = None):
        super().__init__(changes or [])

    def activity_took_place(self, activity: str) -> bool:
        return any(change.activity.used == activity for change in self)

    def __delitem__(self, *args, **kwargs):
        raise TypeError("Cannot delete change from provenance")

    def __setitem__(self, *args, **kwargs):
        raise TypeError("Cannot modify change from provenance")

    def _repr_html_(self) -> str:
        text = "<br /><strong>Provenance</strong>:<ul>"

        for change in self:
            text += f"<li>{change.description}</li>"

        text += "</ul>"

        return text

    def activity(self, id_: URIRef) -> Activity | None:
        return next((change.activity for change in self if change.activity.id_ == id_), None)

    def agent(self, id_: URIRef) -> Agent | None:
        return next((change.agent for change in self if change.agent.id_ == id_), None)

    def target_entity(self, id_: URIRef) -> Entity | None:
        return next(
            (change.target_entity for change in self if change.target_entity.id_ == id_),
            None,
        )

    def source_entity(self, id_: URIRef) -> Entity | None:
        return next(
            (change.source_entity for change in self if change.source_entity.id_ == id_),
            None,
        )

    def as_triples(self) -> Iterable[Triple]:
        for change in self:
            yield from change.as_triples()



import sys
from abc import abstractmethod
from collections.abc import Hashable, Iterator, Sequence
from dataclasses import dataclass
from typing import Any, SupportsIndex, TypeAlias, TypeVar, overload

import pandas as pd
from rdflib import Literal, URIRef

if sys.version_info <= (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

T_ID = TypeVar("T_ID", bound=Hashable)


@dataclass
class NeatObject:
    """A neat object can be dumped to a dictionary."""

    @abstractmethod
    def dump(self, aggregate: bool = True) -> dict[str, Any]:
        """Return a dictionary representation of the object."""
        raise NotImplementedError()

    def _repr_html_(self) -> str:
        return pd.Series(self.dump(aggregate=True)).to_frame(name="value")._repr_html_()


@dataclass(frozen=True)
class FrozenNeatObject:
    """A frozen neat object can be dumped to a dictionary."""

    @abstractmethod
    def dump(self, aggregate: bool = True) -> dict[str, Any]:
        """Return a dictionary representation of the object."""
        raise NotImplementedError()

    def _repr_html_(self) -> str:
        return pd.Series(self.dump(aggregate=True)).to_frame(name="value")._repr_html_()


T_NeatObject = TypeVar("T_NeatObject", bound=NeatObject | FrozenNeatObject)


class NeatList(list, Sequence[T_NeatObject]):
    """A list of neat objects."""

    def dump(self) -> list[dict[str, Any]]:
        """Return a list of dictionary representations of the objects."""
        return [obj.dump() for obj in self]

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame(self.dump())

    def _repr_html_(self) -> str:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]

    # Implemented to get correct type hints
    def __iter__(self) -> Iterator[T_NeatObject]:
        return super().__iter__()

    @overload
    def __getitem__(self, index: SupportsIndex) -> T_NeatObject: ...

    @overload
    def __getitem__(self, index: slice) -> Self: ...

    def __getitem__(self, index: SupportsIndex | slice, /) -> T_NeatObject | Self:
        if isinstance(index, slice):
            return type(self)(super().__getitem__(index))
        return super().__getitem__(index)


Triple: TypeAlias = tuple[URIRef, URIRef, Literal | URIRef]
InstanceType: TypeAlias = URIRef



import math
from collections.abc import Hashable
from typing import TYPE_CHECKING, Any, ClassVar

import pandas as pd
from pydantic import Field, field_serializer, field_validator, model_validator
from pydantic_core.core_schema import SerializationInfo
from rdflib import Namespace, URIRef

from cognite.neat._constants import get_default_prefixes_and_namespaces
from cognite.neat._issues.errors import PropertyDefinitionError
from cognite.neat._rules._constants import EntityTypes
from cognite.neat._rules.models._base_rules import (
    BaseMetadata,
    BaseRules,
    DataModelAspect,
    RoleTypes,
    SheetList,
    SheetRow,
)
from cognite.neat._rules.models._types import (
    ClassEntityType,
    InformationPropertyType,
    MultiValueTypeType,
    URIRefType,
)

# NeatIdType,
from cognite.neat._rules.models.data_types import DataType
from cognite.neat._rules.models.entities import (
    ClassEntity,
    ClassEntityList,
    Entity,
    UnknownEntity,
)

if TYPE_CHECKING:
    from cognite.neat._rules.models import DMSRules


class InformationMetadata(BaseMetadata):
    role: ClassVar[RoleTypes] = RoleTypes.information
    aspect: ClassVar[DataModelAspect] = DataModelAspect.logical

    # Linking to Conceptual and Physical data model aspects
    physical: URIRef | str | None = Field(None, description="Link to the physical data model aspect")
    conceptual: URIRef | str | None = Field(None, description="Link to the conceptual data model aspect")


def _get_metadata(context: Any) -> InformationMetadata | None:
    if isinstance(context, dict) and isinstance(context.get("metadata"), InformationMetadata):
        return context["metadata"]
    return None


class InformationClass(SheetRow):
    """
    Class is a category of things that share a common set of attributes and relationships.

    Args:
        class_: The class ID of the class.
        description: A description of the class.
        implements: Which classes the current class implements.
    """

    class_: ClassEntityType = Field(
        alias="Class", description="Class id being defined, use strongly advise `PascalCase` usage."
    )
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the class.")
    description: str | None = Field(alias="Description", default=None, description="Short description of the class.")
    implements: ClassEntityList | None = Field(
        alias="Implements",
        default=None,
        description="List of classes (comma separated) that the current class implements (parents).",
    )
    instance_source: URIRefType | None = Field(
        alias="Instance Source",
        default=None,
        description="The link to to the rdf.type that have the instances for this class.",
    )
    physical: URIRefType | None = Field(
        None,
        description="Link to the class representation in the physical data model aspect",
    )
    conceptual: URIRefType | None = Field(None, description="Link to the conceptual data model aspect")

    def _identifier(self) -> tuple[Hashable, ...]:
        return (self.class_,)

    @field_serializer("class_", when_used="unless-none")
    def remove_default_prefix(self, value: Any, info: SerializationInfo) -> str:
        if (metadata := _get_metadata(info.context)) and isinstance(value, Entity):
            return value.dump(prefix=metadata.prefix, version=metadata.version)
        return str(value)

    @field_serializer("implements", when_used="unless-none")
    def remove_default_prefixes(self, value: Any, info: SerializationInfo) -> str:
        if isinstance(value, list) and (metadata := _get_metadata(info.context)):
            return ",".join(
                (
                    class_.dump(prefix=metadata.prefix, version=metadata.version)
                    if isinstance(class_, Entity)
                    else str(class_)
                )
                for class_ in value
            )
        return ",".join(str(value) for value in value)


class InformationProperty(SheetRow):
    """
    A property is a characteristic of a class. It is a named attribute of a class that describes a range of values
    or a relationship to another class.

    Args:
        class_: Class ID to which property belongs
        property_: Property ID of the property
        name: Property name.
        value_type: Type of value property will hold (data or link to another class)
        min_count: Minimum count of the property values. Defaults to 0
        max_count: Maximum count of the property values. Defaults to None
        default: Default value of the property
        instance_source: Actual rule for the transformation from source to target representation of
              knowledge graph. Defaults to None (no transformation)
    """

    class_: ClassEntityType = Field(
        alias="Class", description="Class id that the property is defined for, strongly advise `PascalCase` usage."
    )
    property_: InformationPropertyType = Field(
        alias="Property", description="Property id, strongly advised to `camelCase` usage."
    )
    name: str | None = Field(alias="Name", default=None, description="Human readable name of the property.")
    description: str | None = Field(alias="Description", default=None, description="Short description of the property.")
    value_type: DataType | ClassEntityType | MultiValueTypeType | UnknownEntity = Field(
        alias="Value Type",
        union_mode="left_to_right",
        description="Value type that the property can hold. It takes either subset of XSD type or a class defined.",
    )
    min_count: int | None = Field(
        alias="Min Count",
        default=None,
        description="Minimum number of values that the property can hold. "
        "If no value is provided, the default value is  `0`, "
        "which means that the property is optional.",
    )
    max_count: int | float | None = Field(
        alias="Max Count",
        default=None,
        description="Maximum number of values that the property can hold. "
        "If no value is provided, the default value is  `inf`, "
        "which means that the property can hold any number of values (listable).",
    )
    default: Any | None = Field(alias="Default", default=None, description="Default value of the property.")
    instance_source: list[URIRefType] | None = Field(
        alias="Instance Source",
        default=None,
        description="The URIRef(s) in the graph to get the value of the property.",
    )
    inherited: bool = Field(
        default=False,
        exclude=True,
        alias="Inherited",
        description="Flag to indicate if the property is inherited, only use for internal purposes",
    )

    physical: URIRefType | None = Field(
        None,
        description="Link to the class representation in the physical data model aspect",
    )
    conceptual: URIRefType | None = Field(None, description="Link to the conceptual data model aspect")

    def _identifier(self) -> tuple[Hashable, ...]:
        return self.class_, self.property_

    @field_validator("max_count", mode="before")
    def parse_max_count(cls, value: int | float | None) -> int | float | None:
        if value is None:
            return float("inf")
        return value

    @field_validator("instance_source", mode="before")
    def split_on_comma(cls, value: Any) -> Any:
        if isinstance(value, str):
            return [v.strip() for v in value.split(",")]
        return value

    @model_validator(mode="after")
    def set_type_for_default(self):
        if self.type_ == EntityTypes.data_property and self.default:
            default_value = self.default[0] if isinstance(self.default, list) else self.default

            if type(default_value) is not self.value_type.python:
                try:
                    if isinstance(self.default, list):
                        updated_list = []
                        for value in self.default:
                            updated_list.append(self.value_type.python(value))
                        self.default = updated_list
                    else:
                        self.default = self.value_type.python(self.default)

                except Exception:
                    raise PropertyDefinitionError(
                        self.class_,
                        "Class",
                        self.property_,
                        f"Default value {self.default} is not of type {self.value_type.python}",
                    ) from None
        return self

    @field_serializer("instance_source", when_used="unless-none")
    def serialize_instance_source(self, value: list[URIRefType] | None) -> str | None:
        if value is None:
            return None
        return ",".join(str(v) for v in value)

    @field_serializer("max_count", when_used="json-unless-none")
    def serialize_max_count(self, value: int | float | None) -> int | float | None | str:
        if isinstance(value, float) and math.isinf(value):
            return None
        return value

    @field_serializer("class_", "value_type", when_used="unless-none")
    def remove_default_prefix(self, value: Any, info: SerializationInfo) -> str:
        if (metadata := _get_metadata(info.context)) and isinstance(value, Entity):
            return value.dump(prefix=metadata.prefix, version=metadata.version)
        return str(value)

    @property
    def type_(self) -> EntityTypes:
        """Type of property based on value type. Either data (attribute) or object (edge) property."""
        if isinstance(self.value_type, DataType):
            return EntityTypes.data_property
        elif isinstance(self.value_type, ClassEntity):
            return EntityTypes.object_property
        else:
            return EntityTypes.undefined

    @property
    def is_mandatory(self) -> bool:
        """Returns True if property is mandatory."""
        return self.min_count not in {0, None}

    @property
    def is_list(self) -> bool:
        """Returns True if property contains a list of values."""
        return self.max_count in {float("inf"), None} or (
            isinstance(self.max_count, int | float) and self.max_count > 1
        )


class InformationRules(BaseRules):
    metadata: InformationMetadata = Field(alias="Metadata", description="Metadata for the logical data model")
    properties: SheetList[InformationProperty] = Field(alias="Properties", description="List of properties")
    classes: SheetList[InformationClass] = Field(alias="Classes", description="List of classes")
    prefixes: dict[str, Namespace] = Field(
        alias="Prefixes",
        default_factory=get_default_prefixes_and_namespaces,
        description="the definition of the prefixes that are used in the semantic data model",
    )

    @field_validator("prefixes", mode="before")
    def parse_str(cls, values: Any) -> Any:
        if isinstance(values, dict):
            return {key: Namespace(value) if isinstance(value, str) else value for key, value in values.items()}
        elif values is None:
            values = get_default_prefixes_and_namespaces()
        return values

    @model_validator(mode="after")
    def set_neat_id(self) -> "InformationRules":
        namespace = self.metadata.namespace

        for class_ in self.classes:
            if not class_.neatId:
                class_.neatId = namespace[class_.class_.suffix]
        for property_ in self.properties:
            if not property_.neatId:
                property_.neatId = namespace[f"{property_.class_.suffix}/{property_.property_}"]

        return self

    def update_neat_id(self) -> None:
        """Update neat ids"""

        namespace = self.metadata.namespace

        for class_ in self.classes:
            class_.neatId = namespace[class_.class_.suffix]
        for property_ in self.properties:
            property_.neatId = namespace[f"{property_.class_.suffix}/{property_.property_}"]

    def sync_with_dms_rules(self, dms_rules: "DMSRules") -> None:
        # Sync at the metadata level
        if dms_rules.metadata.logical == self.metadata.identifier:
            self.metadata.physical = dms_rules.metadata.identifier
        else:
            # if models are not linked to start with, we skip
            return None

        info_properties_by_neat_id = {prop.neatId: prop for prop in self.properties}
        dms_properties_by_neat_id = {prop.neatId: prop for prop in dms_rules.properties}
        for neat_id, prop in dms_properties_by_neat_id.items():
            if prop.logical in info_properties_by_neat_id:
                info_properties_by_neat_id[prop.logical].physical = neat_id

        info_classes_by_neat_id = {cls.neatId: cls for cls in self.classes}
        dms_views_by_neat_id = {view.neatId: view for view in dms_rules.views}
        for neat_id, view in dms_views_by_neat_id.items():
            if view.logical in info_classes_by_neat_id:
                info_classes_by_neat_id[view.logical].physical = neat_id

    def as_dms_rules(self) -> "DMSRules":
        from cognite.neat._rules.transformers._converters import _InformationRulesConverter

        return _InformationRulesConverter(self).as_dms_rules()

    @classmethod
    def display_type_name(cls) -> str:
        return "VerifiedInformationModel"

    def _repr_html_(self) -> str:
        summary = {
            "type": "Logical Data Model",
            "intended for": "Information Architect",
            "name": self.metadata.name,
            "external_id": self.metadata.external_id,
            "version": self.metadata.version,
            "classes": len(self.classes),
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore
